#ifndef _COMPILE_H_
#define _COMPILE_H_ 1

// Centralise all debug options here:
int debug_phases    = FALSE; // short messages to show which high-level step we're on
int debug_ast       = FALSE; // print the AST tree
int debug_errors    = FALSE; // extra info for programmer on a syntax error
int debug_old_io    = FALSE; // dpsf() etc
int debug_ctuples   = FALSE;
int debug_symbols   = FALSE;
int debug_new_exprs = FALSE;
int debug_cast_creation = FALSE;
int debug_prefetch  = FALSE;

/*static*/ FILE *OUTFILE = NULL, *NULLFILE = NULL;

#include "symtab.h"
#include "datatypes.h"

// These offsets are a poor man's struct:
// Offsets in the A record are different from offsets in the AST.
// Please do not ever write code with absolute offsets!

// A record offsets:
#define A_literal_offset  3

// AST offsets:
// item 0 is reserved.
#define AST_op_offset        1
#define AST_alt_offset       2
#define AST_count_offset     3
#define AST_type_info_offset 4 // experimental new field
#define AST_line_no_offset   5 // previously accessible as child(...,0) - deprecated use sourceline() macro now.

// Any more new common fields (such as a pointer to the source text) would be inserted here.

#define AST_child1_offset 6   // SEE ALSO: ../parser.c:  #define dAST_child1_offset 4
                              //           and compile-default.c

// Adding the offset inconsistently caused one set of errors, but
// even when adding it apparently correctly, these files still fail:

//  ls -l tests/validation-suite/extern.imp
//  ls -l tests/imptoc/rt25a.imp
//  ls -l tests/imp68k/DJRBUG.imp


// Must use these macros for children of the AST, in case new fields are added.
#define AstOP(astp)      AST[(astp)+AST_op_offset              ]
#define Alt0(astp)       AST[(astp)+AST_alt_offset             ]
#define AltCount(astp)   AST[(astp)+AST_count_offset           ]
#define Typefield(astp)  AST[(astp)+AST_type_info_offset       ]
#define sourceline(astp) AST[(astp)+AST_line_no_offset         ]
#define leftchild(astp)  AST[(astp)+AST_child1_offset + ( 1 -1)]
#define rightchild(astp) AST[(astp)+AST_child1_offset + ( 2 -1)]
#define child(astp,xx)   AST[(astp)+AST_child1_offset + ((xx)-1)]
// (see also 'ctuple' and 'detuple' below)

#define CAstOP(x) AstOP(x)

//#ifdef INCLUDE_AST_DEBUG
//#include INCLUDE_AST_DEBUG
//#endif

void warn  (const char *format, ...); /* Returns */
void fault (const char *format, ...); /* Does not return */

void out_inner(int astp, int line);
#define out(astp) out_inner(astp, __LINE__)

int Declare(char *info, symbol *Sym, int line);
#define CFG_Declare(info, sym) Declare(info, sym, __LINE__)

int compile_inner(int astp, int extra, int line); //  __attribute__((warn_unused_result));
#define compile(astp) compile_inner(astp, -1, __LINE__)
#define compile_with_extra(astp,extra) compile_inner(astp, extra, __LINE__)

void unary_op (int astop, int left_astp);
void binary_op(int astop, int left_astp, int right_astp);

// note for later: using 81818181 as the magic marker because we're
// likely to add support for unassigned variable checking later,
// at which point we'll legitimately be using 80808080 and don't
// want the ctuple/detuple macros to break at that point and
// not understand why!
#define ctuple(c_op...) ctuple_inner(__LINE__, c_op, 0x81818181)
int ctuple_inner(int line, int c_op, ...);

// detuple takes a ctuple and unpacks it into individual int * variables.
// Number of variables is checked against the number inserted into the tuple.
// (Would 'unpack' be a better name for this function?)
static int END_MARKER;
#define detuple(astp...) detuple_inner(__LINE__, astp, &END_MARKER)
void detuple_inner(int line, int astp, int c_op, int *ptrs, ...);

void output_condition_C(int astp);

// Low-level routine that actually outputs the generated C:
void C(const char *format, ...);


#define name_idx_to_str(Sym, n) ((char *)&rawstringpool((Sym)->name_idx[n]))
#define pool_to_str(n) ((char *)&rawstringpool(n))

void debug_sym(symbol *Sym, char *info);

// in parser.c:
void indent (int n, FILE * f);

static void RECONSTITUTE(int idx, FILE *out);

#endif
