#ifndef _DATATYPES_H_
#define _DATATYPES_H_ 1

// command-line options governing generation of C code.
int imp_option_trace = 0;  // insert explicit entry/exit tracing, which can also be used for a crude backtrace facility
                           // 0 No tracing
                           // 1 Trace using less detailed GCC mechanism
                           // 2 Build using gdb and use gdb only
                           // 3 Trace using _imp_enter() and _imp_leave() but use macro version of return()
                           // 4 Trace using _imp_enter() and _imp_leave() and insert _imp_leave code in source for returns
                           // 5 As (4) but also gdb support
int imp_option_missing_results = 1;
/*
Descriptors

DEF TAG TEXT TYPE FORM SIZE SPEC PREFIX

This item causes a new descriptor to be generated
and placed in the descriptor area. On creation,
the various fields of the DEF are used to
construct the machine-dependent representation
required for the object.

TAG is an identification which will
be used subsequently to refer to
the descriptor.

TEXT is the source-language identifier
given to the object (a null
string if no identifier was
specified).

TYPE is the type of the object:
GENERAL, INTEGER, REAL, STRING,
RECORD, LABEL, SWITCH, FORMAT.
FORM is one of: SIMPLE, NAME, ROUTINE,
FN, MAP, PRED, ARRAY, NARRAY,
ARRAYN, NARRAYN.

SIZE is either the TAG of the
appropriate record format
descriptor for records, the
maximum length of a string
variable, or the precision of
numerical variables: DEFAULT,
BYTE, SHORT. LONG.
SPEC has the value SPEC or NONE
depending on whether or not the
item is a specification.

PREFIX is one of: NONE, OWN, CONST,
EXTERNAL, SYSTEM, DYNAMIC, PRIM,

PERM or SPECIAL. If SPECIAL is
given there will follow an
implementation-dependent
specification of the properties
of the object (such as that it is
to be a register, for example).



Instruction:  Define <tag> [id] <a> <b> <c>

Effect:       A new tag value is created.

 <tag> defines the tag index which  will  be  used  to
       select the value.

 [id]  specifies the actual identifier associated with
       the described object.  It is a sequence of zero
       or more characters terminated by a comma.  This
       identifier will be used for external linkage if
       necessary  unless  overridden by an Alias. [id]
       will  also  be  used  for  run-time  diagnostic
       information.

 <a>   A two-byte value: <a> = T<<4+F where:
       T = 0 : void
           1 : integer            {qualified by <b>}
           2 : real               {qualified by <b>}
           3 : string             {maximum length <b>}
           4 : record             {format <b>}
           5 : boolean
           6 : set
           7 : 8-bit-enumerated   {format <b>}
           8 : 16-bit-enumerated  {format <b>}
           9 : pointer
          10 : char
       11-15 : undefined          {error}

       F = 0 : void
           1 : simple                  {byte}
           2 : indirect                {bytename}
           3 : general label
           4 : recordformat
           5 : undefined               {error}
           6 : switch
           7 : routine
           8 : function
           9 : map
          10 : predicate
          11 : array                   {array}
          12 : array indirect          {arrayname}
          13 : indirect array          {namearray}
          14 : indirect array indirect {namearrayname}
          15 : undefined               {error}

 <b>   If   T  is  INTEGER  <b>  takes  the  following
       meanings:

       <b> =       1, full range
                   2, range 0..255
                   3, range -32768..32767

       If T is REAL <b> takes the following meanings:

       <b> =       1, normal precision
                   4, double precision

       If T is STRING <b> gives the maximum length  of
       the string.
       If  T  is  RECORD  <b>  gives  the  tag  of the
       corresponding recordformat.
       If T is enumerated <b> gives  the  tag  of  the
       dummy  format  used  to identify the enumerated
       value identifiers.

 <c>   is a two-byte value: U<<5+I<<4+S<<3+X where:
          U is 1 check the object for unassigned
               0 otherwise
          I is 1 if the object is an indirect object,
               0 otherwise
          S is 1 if this is a spec,
               0 otherwise
          X  = 0 :: automatic (stack) allocation
               1 :: own
               2 :: constant
               3 :: external
               4 :: system
               5 :: dynamic
               6 :: primitive
               7 :: permanent

       An indirect object (I=1) differs  from  F=2  in
       that F=2 implies that the actual object created
       will  be  a  pointer  and  will be dereferenced
       whenever used unless explicit action  is  taken
       (e.g.  use  of  Assign-Reference).   If  I=1  a
       pointer will be created (usually as an integer)
       and will be treated as an integer (or  address)
       with no automatic dereferencing taking place.

Notes:        The  tag  values  within  a block should be dense and
 preferably consecutive.   All  tag  values  within  a
 block  must  have values greater than the maximum tag
 value yet defined within the enclosing block.
 Tag definitions remain valid until  the  end  of  the
 enclosing block.
 The  tag values used within a recordformat definition
 must all be zero; the fields of a record are selected
 by their position in the  format,  numbered  starting
 from one.

         x = illegal combination

                           1 1 1 1 1
   F = 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4
      *-----------------------------*
  T = 0   | |x| | |x|x| | |x| | |x| |x| |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  1   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  2   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  3   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  4   |x| | |x| |x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  5   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  6   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  7   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  8   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
  9   |x| | |x|x|x|x|x| | |x| | | | |
      |-+-+-+-+-+-+-+-+-+-+-+-+-+-+-|
 10   |x| | |x|x|x|x|x| | |x| | | | |
      *-----------------------------*
 */

/*
F = 0 : void
1 : simple                  {byte}
2 : indirect                {bytename}
3 : general label
4 : recordformat
5 : undefined               {error}
6 : switch
7 : routine
8 : function
9 : map
10 : predicate
11 : array                   {array}
12 : array indirect          {arrayname}
13 : indirect array          {namearray}
14 : indirect array indirect {namearrayname}
15 : undefined               {error}
*/

typedef enum FORM {
  simple, name, _label, recordformat, _switch, routine, function, map, predicate, array, arrayname, namearray, namearrayname
} FORM;
const char *Form[] = {
  "", "name", "label", "recordformat", "switch", "routine", "function", "map", "predicate", "array", "arrayname", "namearray", "namearrayname"
};

typedef enum TYPE {
  no_return_type, general, integer, real, string, record
} TYPE;
const char *Type[] = {
  "", "general", "integer", "real", "string", "record"
};

typedef enum PRECISION {
  prec_not_relevant, byte, _short, word, _long, longlong
} PRECISION;
const char *Precision[] = {
  "", "byte", "short", "word", "long", "longlong"
};

// added _fpp as a special class for parameters to procedures.
typedef enum CLASS {   // Storage class
  badclass, _auto,   own,   constant,   external,  _system,   dynamic,   prim,   perm, _fpp
} CLASS;
const char *Class[] = {
  "?",      "auto", "own", "constant", "external", "system", "dynamic", "prim", "perm"
};

typedef enum X {
  has_data, has_cache, is_spec, has_body, is_anonymous, is_a_formal_param, x_last_item
} X;
const char *Xx[] = {
  "HAS_data", "HAS_cache", "IS_spec", "HAS_body", "IS_anonymous", "IS_formal_param"
};
#define X_HAS_DATA          (1<<has_data)
#define X_HAS_CACHE         (1<<hash_cache)
#define X_IS_SPEC           (1<<is_spec)
#define X_HAS_BODY          (1<<has_body)
#define X_IS_ANONYMOUS      (1<<is_anonymous)
#define X_IS_A_FORMAL_PARAM (1<<is_a_formal_param)
/*

OK, first here is how type information *should* work:

The struct below describes the type information that is associated with a variable, or an array element or
a field in a record etc.  It is *only* the type information used to access the data relative to some external
base address.  It does *not* contain a variable's name etc.  The actual variable itself - the thing that
is entered in the name tables - is a triple of:  name, type information, and address (if needed for a code
generator).  This is not a name table entry, this is a type table entry.  The name table is a separate
structure managed by "symtab.c".

The type information is hierarchical.  When code generating, there will be a top-level type structure
corresponding to say an array.  When an array element is accessed, it will be associated with the type
structure for a member of that array, which might be a record (aka struct).  Access to a field within
that struct is via another type structure.  Etc.  As the compile procedure walks down the tree describing
a variable access, so it also walks down the tree of type structures in parallel, associating the type
information with the operand at each stage.

Consequently it should be clear that the type structure describing where we are in the evaluation of
an operand must allow for anonymous items (eg in the transition from a pointer to a variable to the 
pointed object)

BUT secondly:

during development this is being overloaded as a symbol table manager and includes names where
names might not be necessary.

The issue is complicated somewhat by the Imp compiler trying to put all the info in a single
entry, eg a record namearrayname

whereas the C style is

  name (pointer to)
    array
      name (type of element of array)
        struct (object pointed to by name above)
          struct field 1 (basic type)
          struct field 2
            info for field if not a basic type
          struct field ...

To make this code more generally useful, I am tempted to use the more general C style
of type information for Imp as well...

I'm really in two minds as to a) go ahead with imp style *and* names embedded in the symbol
*or* b) do the rewrite now, which will significantly delay getting any output.

Another 'during development' issue is handling namelists as opposed to a single declaration name.
We may not have all the info for a declaration at the point that the <rest of namelist> is hit,
so for convenience it looks like it may be easier to just stuff them into the struct and then
split them off for the symtab entry once the struct is complete and finalised.
 */

#define MAX_CHILDREN 32
#define MAX_NAMES    128 // How many variables is someone likely to put in a single declaration statement anyway???
#define MAX_DIMS     6

typedef struct symbol symbol;
typedef struct symbol {
  int sequence; // for debugging only.
  // this is a little more verbose that PSR's version.  We have lots of RAM available.
  FORM      form;         // simple, name, _label, recordformat, _switch, routine, function, map, predicate, array, arrayname, namearray, namearrayname
  TYPE      type;         // general, integer, real, string, record
  PRECISION precision;    // base variable width: byte, _short, word, _long, longlong - not if record or string or array
  CLASS     class;        // storage class: _auto, own, constant, external, _system, dynamic, prim, perm
  int       x;            // arbitrary extra bit flags.
                          // So far: X_HAS_DATA X_HAS_CACHE X_IS_SPEC X_HAS_BODY X_IS_ANONYMOUS

  // names extended to an array for now, to handle namelists. Probably should be a linked list - this is a short term expedient.
  int next_name;
  int name_ast[MAX_NAMES];           // variable name as an ast pointer.  This might be a namelist of multiple names all sharing the same declaration details
  int name_idx[MAX_NAMES];           // variable name as a stringpool index. 
  int alias_idx[MAX_NAMES];          // external linkage string

  int string_len_ast;

  int cache_idx;          // cached value of variable for constant folding or optimisation. Also used by const variables

  int param_list_ast;     // Procedure formal parameters. Need a way for compile() to add param to formal parameter list in current Sym. AST for now.  IDX later.
  int code_ast;           // AST item containing body of code

  int optional_data_idx;  // index into array specific to datatype with extra information, for example string length
  int recordformatid_ast; // If this is a structure, then index of structure layout (generated AST object) ie fields and variants
  int array_dimen;        // If this is an array, then index of array layout (generated AST object) ie bounds and dimensions
  int lowerbound_ast[MAX_DIMS];
  int upperbound_ast[MAX_DIMS];
  int init_ast;           // initialisation data for this object
  int scope_idx;          // scope that contains this declaration - needed to determine hierarchy eg module/procname/anon-block23/procname - for code gen and diags

  int children;
  symbol *child[MAX_CHILDREN];        // FLEX ARRAY OF CHILDREN. Eg struct fields. Array is truncated when object is complete.

} symbol;

#endif
