// DO NOT EDIT THIS FILE.
// This code has been re-inserted into compile.c (from where it was originally extracted) to
// make it easier to edit the code generation without switching back and forth between files.

/*
    replacement code and data structures for Imp expression evaluation...

    ctuples for an operator are a pair of <C_ and descriptor> - the descriptor
    may be a declaration symbol but it could also be the same information for
    an expression, constructed bottom-up, to allow type-checking at the top,
    or appropriate casting for assignments/jam transfers etc.


    Calls to:
       ctuple(C_FUNCTION_CALL, name, param)
    are now being replaced by calls to:
       build_call(C_FUNCTION_CALL, name, param)
    which modifies the param list as the function call AST is built, depending on whether address or value is wanted.


What we should be doing... a <NAME> should never make it through to the finished CAST.
A name in the AST should be evaluated by one of these:

  Evaluate_NAME:

    Evaluate_Expression

    Evaluate_Value

    Evaluate_Address

    Evaluate_RT_Call(RT)

    Evaluate_Fn_or_Pred_Call(FN)

    Evaluate_Map_Call(MAP)

    Evaluate_Array_Element(RA)

    Evaluate_Array_NameElement(RA)

    Evaluate_ArrayName(RA)

    Evaluate_Parameters(FPP, APP) (linked list of formal and actual parameters which would be stepped through together)

    Evaluate_Value_Assignment(Dest, Source, assignment_type)

    Evaluate_Address_Assignment(Dest, Source, assignment_type)

    Evaluate_Fn_or_Pred_Result(Source, assignment_type)

    Evaluate_Map_Result(Source)

    Evaluate_Procedure_as_a_parameter

 */

#define CAstOP(x) AstOP(x)

#ifdef NEVER

int debug_new_exprs = 1;

char *CAstOPName(int castop) {
  return CAST[castop-C_BASE].name;
}
#endif

char *AstOPName(int astop) {
  return phrasename[astop];
}

void debug_record_type(int astp, int depth);

void output_c_expression(int castp) {
  if (debug_new_exprs) fprintf(stderr, "output_c_expression: @C_AST:%d\n", castp);
  if (castp < 0) {
    return;
  }
  int castop = CAstOP(castp);
  switch (castop) {
  default:
    fprintf(stderr, "MISSING CAST OPERATOR C_%s IN output_c_expression: @C_AST:%d\n",
            CAstOPName(castop),
            castp);
    return;
  }
}

int samename(int name1, int name2) {
  char *names1, *names2;
  if (name1 == name2) {
    fprintf(stderr, "%s vs %s - fast match\n", name_to_cstr(name1),name_to_cstr(name2));
    return 1;
  }
  names1 = name_to_cstr(name1); names2 = name_to_cstr(name2);
  if (strcmp(names1,names2) == 0) {
    fprintf(stderr, "%s vs %s - match\n", names1,names2);
    return 1;
  } else {
    fprintf(stderr, "%s vs %s - no match\n", names1,names2);
    return 0;
  }
}

int same(int astp1, int astp2) {
  // Initially just check it is a simple type such as C_TYPE_INT and the names match.
  // Later tweaks can handle more complex expressions.
  // Note must return FALSE if it is a proc/fn/map with side-effects.
  if (AstOP(astp1) == C_NAME && AstOP(astp2) == C_NAME) {
    // just the text name, no extra info.  #1 = stringpool index, #2 = astp of <NAME> or perhaps -1.
    if (leftchild(astp1) == -1 || leftchild(astp2) == -1) {
      if (rightchild(astp1) == -1 || rightchild(astp2) == -1) return FALSE;
      return rightchild(astp1) == rightchild(astp2);
    } else {
      if (leftchild(astp1) == leftchild(astp2)) return TRUE; // assuming stringpool works
      //fprintf(stderr, "Compare %s vs %s\n", pool_to_str(leftchild(astp1)), pool_to_str(leftchild(astp2)));
      if (strcmp(pool_to_str(leftchild(astp1)), pool_to_str(leftchild(astp2))) == 0) return TRUE; // assuming stringpool doesn't work :-(
    }
  }
  return FALSE;
}

int build_call(int op, int name, int param) {
#ifdef NEVER
  // We need to get the type information for the parameters to procedure 'name'
  // and modify the param list depending on whether the parameter takes a value or a %name
  //
  char *names = name_to_cstr(name);
  int decl_c = lookup("Declarations", names);
  // #1 name  #2 spare  #3 parameter list (C_TYPE_PARAMETERS)  #4 body  #5 alias or -1
  int name2, result_type, fpp, spec_or_body, alias, sex, linkage;
  int storage_class, linkage_class;
  
  if (AstOP(decl_c) == C_TYPE_PROCEDURE ||
      AstOP(decl_c) == C_TYPE_FN ||
      AstOP(decl_c) == C_TYPE_MAP ||
      AstOP(decl_c) == C_TYPE_PREDICATE)
    detuple(decl_c, AstOP(decl_c), &name2, &result_type, &fpp, &spec_or_body, &alias, &sex);
  else {
    diagnose(decl_c);
    fault("build_call: passed a bad CAST type.");
  }
  assert(AstOP(name2) == C_NAME);
  fprintf(stderr, "Build %s call: type=%s name=%s/%s\n", CAstOPName(AstOP(decl_c)), CAstOPName(op), names,name_to_cstr(name2));
  // Now we loop though fpp and match up the actual parameters...
  while (fpp != -1) {
    int thisparam, paramname; // not used to generate code, but we'll use it in diagnostics
    detuple(fpp, C_TYPE_PARAMETERS, &paramname, &thisparam, &fpp);
    fprintf(stderr, "   param %s: \n",paramname == -1 ? "" : name_to_cstr(paramname));
    diagnose(thisparam);
  }
#endif
  return ctuple(op, name, param);
}

int cform_assignment(int astp) {
  int left1, left2, right, constright, newop, newop2;
  detuple(astp, C_ASSIGN_VALUE, &left1, &right);
  if (AstOP(right) == C_BINPLUS) {
    detuple(right, C_BINPLUS, &left2, &right);
    newop = C_ASSIGN_PLUS; newop2 = C_ASSIGN_PLUSPLUS; 
  } else if (AstOP(right) == C_BINMINUS) {
    detuple(right, C_BINMINUS, &left2, &right);
    newop = C_ASSIGN_MINUS; newop2 = C_ASSIGN_MINUSMINUS;
  } else return astp;
  if (same(left1,left2)) {
    if (AstOP(right) == C_BRACKET) right = leftchild(right);
    constright = IntConstant_c(right);
    if (constright != -1 && AstOP(constright) == C_CONST_INT && child(constright,2) == 1) { // #1 string rep  #2 int val (32 bits)
      return ctuple(newop2, left1);
    }
    return ctuple(newop, left1, right); // if right is redundantly bracketed, safe to remove the backets
  } else if (same(left1,right)) {
    if (AstOP(left2) == C_BRACKET) left2 = leftchild(left2);
    constright = IntConstant_c(left2);
    if (constright != -1 && AstOP(constright) == C_CONST_INT && child(constright,2) == 1) {
      return ctuple(newop2, left1);
    }
    return ctuple(newop, left1, left2); // if right is a const int == 1 safe to mod to PLUSPLUS/MINUSMINUS (newop2)
  }
  return astp;
}

#define VALUE_WANTED 0
#define ADDRESS_WANTED 1

int recompile_record_field(int recordtype_c, int list_of_fields, int field_c, int is_pointer) {
  int i;
  int rfname, first_decln, this_decl, more_decls;
  
  if (list_of_fields != -1) {
    assert(AstOP(list_of_fields) == C_TYPE_RECORD_FIELD);
    this_decl = list_of_fields;
    for (;;) {
      int field_name, field_type, percent_name, next_field;;
      detuple(this_decl, C_TYPE_RECORD_FIELD, &field_name, NULL, &field_type, &percent_name, &next_field);
      //for (i = 0; i < depth*3; i++) fprintf(stderr, " ");

      fprintf(stderr, "  %s %s%s\n", name_to_cstr(field_name), percent_name==-1?"":"*", CAstOPName(AstOP(field_type)));

      if (samename(field_name, rightchild(field_c))) {
        // build a ctuple with '->' or '.' and the field.
        return ctuple(C_RECORD_FIELD, recordtype_c, field_c, is_pointer); // still do do: stuff following record field...
      }
      
      if (AstOP(field_type) == C_TYPE_RECORDFORMAT) {
        // Follow actual records, *not* recordnames...
        //if (percent_name == -1) debug_record_type(field_type, depth+1);
      }
      if (next_field == -1) break;
      // looks like there aren't any more.
      // more_decls was a list of C_SEQ items containing a single C_TYPE_RECORD_FIELD in each,
      // when in fact it should have been built using a C_TYPE_RECORD_FIELD list instead.
    }
    warn("Field not present in record format");
    return -1; // not found!
  }
  /// otherwise we need to locate the list of fields from the record declaration:

  detuple(recordtype_c, C_TYPE_RECORDFORMAT, NULL, &rfname, &first_decln); // #1 <record name or -1>  #2 record format name  #3 field list of C_TYPE_RECORD_FIELD
  //for (i = 0; i < depth*3; i++) fprintf(stderr, " ");
  //fprintf(stderr, "Record: %s  depth=%d first_decln=%d\n", name_to_cstr(rfname), depth, first_decln);
  more_decls = first_decln;
  while (AstOP(more_decls) == C_SEQ) {
    detuple(more_decls, C_SEQ, &this_decl, &more_decls);
    //for (i = 0; i < depth*3; i++) fprintf(stderr, " ");
    fprintf(stderr, " %s\n", CAstOPName(AstOP(this_decl)));
    if (AstOP(this_decl) != C_TYPE_RECORD_FIELD) {
      diagnose(this_decl); // -> C_TYPE_RECORDFORMAT  why?
      return -1;
      //assert(AstOP(this_decl) == C_TYPE_RECORD_FIELD);
    }
    // C_TYPE_RECORD_FIELD =
    // #1 field name
    // #2 field offset if known relative to start of current record, or -1 (for later)
    // #3 type: field type could be a basic type, or a C_TYPE_RECORD_VARIANT list
    // #4 %name?
    // #5 next <C_TYPE_RECORD_FIELD> or -1
    for (;;) {
      int field_name, field_type, percent_name, next_field;;
      detuple(this_decl, C_TYPE_RECORD_FIELD, &field_name, NULL, &field_type, &percent_name, &next_field);
      //for (i = 0; i < depth*3; i++) fprintf(stderr, " ");

      fprintf(stderr, "  %s %s%s\n", name_to_cstr(field_name), percent_name==-1?"":"*", CAstOPName(AstOP(field_type)));

      if (samename(field_name, rightchild(field_c))) {
        // build a ctuple with '->' or '.' and the field.
        return ctuple(C_RECORD_FIELD, recordtype_c, field_c, is_pointer); // still do do: stuff following record field...
      }
      
      if (AstOP(field_type) == C_TYPE_RECORDFORMAT) {
        // Follow actual records, *not* recordnames...
        //if (percent_name == -1) debug_record_type(field_type, depth+1);
      }
      if (next_field == -1) break;
      // looks like there aren't any more.
      // more_decls was a list of C_SEQ items containing a single C_TYPE_RECORD_FIELD in each,
      // when in fact it should have been built using a C_TYPE_RECORD_FIELD list instead.
    }
    if (more_decls == -1) break;
  }
}


#define compile_expression(astp, extra_parameter, address_wanted) compile_expression_inner(astp, extra_parameter, address_wanted, __LINE__)
int compile_expression_inner(int astp, int extra_parameter, int address_wanted, int line) {
  if (debug_new_exprs) fprintf(stderr, "compile_expression(@AST:%d, %d, %s) from line %d\n", astp, extra_parameter, (address_wanted ? "GET ADDRESS" : "GET VALUE"), line);
  if (astp < 0) {
    fprintf(stderr, "compile_expression: -1 from line %d\n", line);
    return -1;
  }
  int value_wanted = !address_wanted;
  int astop = AstOP(astp);
  int alt      = Alt0(astp);
  int count    = AST[astp+AST_count_offset];
  int *A_child = &AST[astp+AST_child1_offset]; // An index into a child of the AREC, *not* a child of the AST.
                                               // (for printing terminals with RECONSTITUTE, although it may be better
                                               //  to pass astp and let RECONSTITUTE apply AST_child1_offset)
  if (astop >= C_BASE) {
    fprintf(stderr, "? redundant compile_expression(C_%s,%d,%d) called from line %d\n", CAstOPName(astop), 0, VALUE_WANTED, line);
    return astp; // already compiled.
  }

  switch (astop) {

    case AST_BASIC_UI:           //\\ P<BASIC-UI> = 
      if (alt == 0) {            //\\   <LVALUE> <OPT-ASSIGN>;
        int lvalue_c;
        int assign_c    = compile(child(astp,2));

        if (assign_c == -1) {
          assign_c    = compile_expression(child(astp,1), 0, VALUE_WANTED);
        } else if (AstOP(assign_c) == C_ASSIGN_VALUE) {
          leftchild(assign_c) = compile_expression(child(astp,1), 0, VALUE_WANTED);
          assign_c = cform_assignment(assign_c);
        } else if (AstOP(assign_c) == C_ASSIGN_ADDRESS) {
          leftchild(assign_c) = compile_expression(child(astp,1), 0, ADDRESS_WANTED); // ptr == var
        } else if (AstOP(assign_c) == C_STR_RES) {
          leftchild(assign_c) = compile_expression(child(astp,1), 0, VALUE_WANTED);
        } else if (AstOP(assign_c) == C_JAM_TRANSFER) {
          leftchild(assign_c) = compile_expression(child(astp,1), 0, VALUE_WANTED);
        } else {
          warn("Assignment is of unexpected type %s", CAstOPName(AstOP(assign_c)));
        }
        return ctuple(C_ADD_SEMI, assign_c);
      }

    case AST_OPT_ASSIGN:         //\\ P<OPT-ASSIGN> =
      if (alt == 0) {            //\\   '-' '>' <STRING-RESOLUTION-EXPR>,
        return ctuple(C_STR_RES, -1, compile_expression(child(astp,3), 0, VALUE_WANTED));
        
      } else if (alt == 1) {     //\\   <ASSOP-EXPR>,
        int assop_expr = compile_expression(leftchild(astp), 0, VALUE_WANTED);
        assert(AstOP(assop_expr == C_ASSIGN_ADDRESS || AstOP(assop_expr) == C_ASSIGN_VALUE || AstOP(assop_expr) == C_JAM_TRANSFER));
        assert(leftchild(assop_expr) == -1);
        return assop_expr; // FOUND YOU, YOU BASTARD!

      } else if (alt == 2) {     //\\   ;
        return -1;
        
      }

      // These two parse identically but subfields are separated out from simple fields so they can be evaluated differently:
    case AST_RECORDFIELD:        //\\ P<RECORDFIELD> =
                                 //\\   <NAME> <Opt-PARAMETERS-OR-POINTERINDEX> <Opt-RECORDFIELD>;
    case AST_SUBFIELD:           //\\ P<SUBFIELD> =
                                 //\\   <NAME> <Opt-PARAMETERS-OR-POINTERINDEX> <Opt-SUBFIELD>;
    case AST_LVALUE:             //\\ P<LVALUE> =
      if (alt == 0) {            //\\   <NAME> <Opt-PARAMETERS-OR-POINTERINDEX> <Opt-RECORDFIELD>;

        // I realised rather late that recordfields have to be right-associative, i.e. the top of the tree is
        // the final recordfield node in a tree of recordfield nodes.  This is because the type of the whole
        // expression is governed by the type of the final recordfield.

        // It's easy enough to build the tree that way, but then the complication becomes the leftmost part
        // of the variable which may be a function or map and will still require the '*' indirection or
        // not depending on the assignment type.  Also the address of this conglomeration is the address
        // of the rightmost element regardless of how we got there... So although the '&' is applied to the
        // top-level, we don't know whether it is needed or not until we get all the way to the bottom, eg
        // with record_next_int vs record_next_intname

        int name_c    = compile_expression(child(astp,1), 0, VALUE_WANTED);
        assert(AstOP(name_c) == C_NAME);
        int lvalue_c = name_c;
        char *names = name_to_cstr(lvalue_c);
        int lvalue_decl = lookup("Declarations", names);
        int ispointer;
        // (top level) RECORD FIELD:

HACKY_RESTART: // come back here if the object's type information is behind a C_TYPE_PARAMETERS tuple.

        ispointer = FALSE; // SHOULD reflect whether lvalue_c is a pointer to a struct or a struct itself.

        if (lvalue_decl == -1) {
          // Should not get here if evaluating a record field name.  Should be handled by recompile_record_field()
          warn("Undeclared variable %s", names); // make this a 'fault' later, after all the test suite programs have declared the variables they use :-)
          return name_c;                         // undeclared variable - we'll just pass 'name' through unaltered and ignore any params, indexes or subfields.
        }

        int param_c   = compile_expression(child(astp,2), 0, VALUE_WANTED);
                        // could be a map parameter list *or* array indexes. They look the same so need type info to disambiguate.
                        // C_PTRIDX or C_IMPPLUSPLUS or C_IMPMINUSMINUS or C_PARAMETER_LIST or -1 (all -1, ...)

        int field_ast = child(astp,3); // not yet compiled. We need to know context in order to compile (ie address_wanted or value_wanted?)
        int field_c   = -1;

        if ((param_c == -1) && (Alt0(field_ast) == 1)) {  // remember, not yet compiled, so we're looking at the parse alternatives, not the returned ctuple
          // This *can* fall out in the wash by following the path below, but we'll
          // special-case simple variables for speed and simplicity...
          // could be fn or map ...
          if (AstOP(lvalue_decl) == C_DECLARE_SCALAR) {
            // #1 name  #2  %sex  #3  pointer to(ie %name) (or -1)  #4  arrayname  #5 C_TYPE...  #6 initialisation
            // perform type-checking here. output procedure doesn't care, just wants the name
            if (address_wanted) { // BUG: see tests/test191.imp and decide if intname == lvalue is address_wanted or value_wanted on LHS...
              if (child(lvalue_decl,3) != -1) return name_c; // it's already a %name
              return ctuple(C_DEREFERENCE_NAME, name_c);
            }
            if (child(lvalue_decl,3) != -1) return ctuple(C_DEREFERENCE_NAME, name_c); // value of a %name wanted
            return name_c;
          } else if (AstOP(lvalue_decl) == C_BRACKETED_LIST) {
            fault("C_BRACKETED_LIST not yet handled.");
            return name_c;
          } else if (AstOP(lvalue_decl) == C_PARAMETER_LIST) {
            fault("C_PARAMETER_LIST not yet handled.");
            return name_c;
          } else if (AstOP(lvalue_decl) == C_INDEX_LIST) {
            fault("C_INDEX_LIST not yet handled.");
            return name_c;
          } else if (AstOP(lvalue_decl) == C_TYPE_PROCEDURE ||
                     AstOP(lvalue_decl) == C_TYPE_FN ||
                     AstOP(lvalue_decl) == C_TYPE_MAP ||
                     AstOP(lvalue_decl) == C_TYPE_PREDICATE) {          
            // #1 name #2 spare       #3 parameter list (C_TYPE_PARAMETERS)  #4 body or spec or -1  #5 alias or -1  #6 %sex
            lvalue_c = build_call(C_FUNCTION_CALL, lvalue_c, ctuple(C_PARAMETER_LIST, lvalue_c, -1, -1));
            if (address_wanted) {
              if (AstOP(lvalue_decl) == C_TYPE_MAP) return lvalue_c; // it's already a %name
              // What we haven't yet handled properly is passing a procedure name as a parameter
              // which C allows to have a '&' in front of the name but doesn't insist on it.
              // But we should not interpret one of these as a *call* to the procedure.
              return ctuple(C_DEREFERENCE_NAME, lvalue_c);
            }
            if (AstOP(lvalue_decl) == C_TYPE_MAP) return ctuple(C_DEREFERENCE_NAME, lvalue_c); // value of a %name wanted
            return lvalue_c;
          } else if (AstOP(lvalue_decl) == C_DECLARE_ARRAY) {
            warn("C_DECLARE_ARRAY not yet handled.");
            // I don't think a DECLARE_* tuple should find its way here. DECLAREs should only be used
            // to output the actual declaration.  The type information should be separately stored in a C_TYPE_* tuple.
            // So remove this branch later if it never happens.
            /* (btw What would evaluate a bare arrayname without attached indexing
                    is arrayname == array, or passing an array as a parameter) */
            return name_c;
          } else if (AstOP(lvalue_decl) == C_TYPE_PARAMETERS) {
            lvalue_decl = child(lvalue_decl,2);
            warn("C_TYPE_PARAMETERS not yet handled - retrying as a %s", CAstOPName(AstOP(lvalue_decl)));
            // #1 name  #2 C_TYPE_...  #3 next <C_TYPE_PARAMETERS> or -1
            goto HACKY_RESTART;

          } else if (AstOP(lvalue_decl) == C_TYPE_INT) {
            return name_c;
          } else if (AstOP(lvalue_decl) == C_TYPE_POINTER_TO) {
            if (value_wanted) return ctuple(C_DEREFERENCE_NAME, name_c); else return name_c;
          } else {
            warn("Simple variable type info for %s", names);
            diagnose(lvalue_decl);
          }
          if (value_wanted) return name_c;
          return ctuple(C_DEREFERENCE_NAME, name_c);  // Should I be returning, instead, a ctuple(C_OBJECT, name_c, lvalue_decl) ???
                                                             // I.e. a new construct of both value and type information???
        }
        
        if (param_c != -1) {
          // C_PTRIDX or C_IMPPLUSPLUS or C_IMPMINUSMINUS or C_PARAMETER_LIST (etc) or -1 (all -1, ...)
          // Either a fn/map or an array access...

          if (leftchild(param_c) != -1) {
            fprintf(stderr, "Inserting a %s record as the parent of a %s.\n", CAstOPName(AstOP(lvalue_c)), CAstOPName(AstOP(param_c)));
            fprintf(stderr, "unfortunately the leftchild(param_c) is already set to:");
            diagnose(param_c);
            assert(leftchild(param_c) == -1);
          }

          if (AstOP(param_c) == C_PTRIDX) {
              // TO DO!
              warn("Imp 68K 'ptr[offset]' pointer indexing not yet implemented");
              return name_c;
          } else if (AstOP(param_c) == C_IMPPLUSPLUS) {
              // TO DO!
              warn("Imp 68K 'ptr++offset' pointer indexing not yet implemented");
              return name_c;
          } else if (AstOP(param_c) == C_IMPMINUSMINUS) {
              // TO DO!
              warn("Imp 68K 'ptr--offset' pointer indexing not yet implemented");
              return name_c;
          } else if ( AstOP(param_c) == C_PARAMETER_LIST ||
                      AstOP(param_c) == C_BRACKETED_LIST ||
                      AstOP(param_c) == C_INDEX_LIST
                      ) {
            
            if (AstOP(lvalue_decl) == C_TYPE_FN) {
              AstOP(param_c) = C_PARAMETER_LIST;
              lvalue_c = build_call(C_FUNCTION_CALL, lvalue_c, param_c);
              // fn and params: #1 = proc name (C_NAME), #2 = C_PARAMETER_LIST
              // #1 object this hangs off  #2 actual parameter  #3 rest of actual parameter list (another C_PARAMETER_LIST)
              if (address_wanted) warn("Cannot take address of %%function %s", names);
              if (Alt0(field_ast) == 1) { // remember, not yet compiled, so we're looking at the parse alternatives, not the returned ctuple
                return lvalue_c;
              }
            } else if (AstOP(lvalue_decl) == C_TYPE_PREDICATE) {
              AstOP(param_c) = C_PARAMETER_LIST;
              lvalue_c = build_call(C_FUNCTION_CALL, lvalue_c, param_c);
              // fn and params: #1 = proc name (C_NAME), #2 = C_PARAMETER_LIST
              // #1 object this hangs off  #2 actual parameter  #3 rest of actual parameter list (another C_PARAMETER_LIST)
              if (address_wanted) warn("Cannot take address of %%predicate %s", names);
              if (Alt0(field_ast) == 1) { // remember, not yet compiled, so we're looking at the parse alternatives, not the returned ctuple
                return lvalue_c;
              }
              
            } else if (AstOP(lvalue_decl) == C_TYPE_PROCEDURE) {
              AstOP(param_c) = C_PARAMETER_LIST;
              lvalue_c = build_call(C_FUNCTION_CALL, lvalue_c, param_c);
              // #1 name #2 spare       #3 parameter list (C_PARAMETER_LIST)  #4 body or spec or -1  #5 alias or -1  #6 %sex
              // #1 object this hangs off  #2 actual parameter  #3 rest of actual parameter list (another C_PARAMETER_LIST)
              if (address_wanted) warn("Cannot take address of %%routine %s", names);
              if (Alt0(field_ast) == 1) { // remember, not yet compiled, so we're looking at the parse alternatives, not the returned ctuple
                return lvalue_c;
              }
              
            } else if (AstOP(lvalue_decl) == C_TYPE_MAP) {
              AstOP(param_c) = C_PARAMETER_LIST;
              lvalue_c = build_call(C_FUNCTION_CALL, lvalue_c, param_c);
              // map with no parameters
              // fn and params: #1 = proc name (C_NAME), #2 = C_PARAMETER_LIST
              // #1 object this hangs off  #2 actual parameter  #3 rest of actual parameter list (another C_PARAMETER_LIST)

#ifdef NEVER
              if (value_wanted) {
                AstOP(lvalue_decl) = C_TYPE_FN; // subtle hack. I'll explain later.
              } else {
                param_c = ctuple(C_DEREFERENCE_NAME, lvalue_c);
              }
#else
              // Needs work.  Sort out the ASSIGN_* stuff first.
              if (value_wanted) lvalue_c = ctuple(C_DEREFERENCE_NAME, lvalue_c);
#endif
              if (Alt0(field_ast) == 1) { // remember, not yet compiled, so we're looking at the parse alternatives, not the returned ctuple
                return lvalue_c;
              }

            } else if (AstOP(lvalue_decl) == C_DECLARE_SCALAR) {
              // test for some kind of %name
              int pointer_to, arrayname;
              detuple(lvalue_decl, C_DECLARE_SCALAR, NULL, NULL, &pointer_to, &arrayname, NULL, NULL);
              if (pointer_to != -1) lvalue_c = ctuple(C_DEREFERENCE_NAME, lvalue_c);
              
            } else if (AstOP(lvalue_decl) == C_DECLARE_ARRAY) {
              AstOP(param_c) = C_INDEX_LIST;

            } else if (AstOP(lvalue_decl) == C_TYPE_PARAMETERS) {
              warn("Parameter %s is not handled.", names);
              return name_c;              

            } else {
              warn("Type of object (%s) for %s(...) not handled.", CAstOPName(AstOP(lvalue_decl)), names);
              return name_c;
            }
            leftchild(param_c) = lvalue_c; // first field of all 4 options is the object they hang off:
                                           // C_PTRIDX or C_IMPPLUSPLUS or C_IMPMINUSMINUS or C_PARAMETER_LIST
          } else {
            warn("Unexpected AstOP when evaluating an lvalue: %s", CAstOPName(AstOP(param_c)));
            return name_c; // temporary.
          }
          lvalue_c = /* updated */ param_c;
        }

        if (Alt0(field_ast) == 1) { // remember, not yet compiled, so we're looking at the parse alternatives, not the returned ctuple
          // NOT A RECORD WITH A FIELD.
          return lvalue_c;
        }

        // NOW SUBFIELDS
        
        if (AstOP(lvalue_decl) == C_DECLARE_SCALAR) {
          if (child(lvalue_decl,3) != -1) ispointer = TRUE;
        }
        // (Need another test for when lvalue is an array element)

        field_c = compile_expression(child(astp,3), 0, VALUE_WANTED); // C_RECORD_FIELD (-1(parent), field, -1 (ispointer?)) or -1
        //                                          ^ This is where the type information needs to be inserted for subfields.

        // #1 parent  #2 child  #3 is pointer? ('->' vs '.')
        child(field_c,1) = lvalue_c; // Need a recursive call using the type information for this record field in order to get elements which are array members.
        child(field_c,3) = ispointer;
        
        /*
        fprintf(stderr, "lvalue_decl: "); diagnose(lvalue_decl);
        fprintf(stderr, "\nlvalue_c:    "); diagnose(lvalue_c);
        fprintf(stderr, "\nfield_c:     "); diagnose(field_c);
        fprintf(stderr, "\n");
        */
        return field_c;
      }
      
    //\\ #  These pointer arithmetic elements are for Hamish's PDP9/15 Imp (old): INDP == INDEX0++P
    //\\ #  and for Hamish's 68000 Imp:     %if tp[1]_mode # litmode %then lo = minint %else lo = tp[1]_val
    //\\ #  I hope this grammar definition causes the right precedence level...
    //\\ # 
    //\\ #  They can also be used in LVALUEs.
    case AST_Opt_PARAMETERS_OR_POINTERINDEX://\\ P<Opt-PARAMETERS-OR-POINTERINDEX> =
      if (alt == 0) {            //\\   <immediate:return !in_const_initialiser_list;> '[' <EXPR> ']',
        return ctuple(C_PTRIDX, -1, compile_expression(child(astp,3), 0, VALUE_WANTED));
      } else if (alt == 1) {     //\\   <immediate:return !in_const_initialiser_list;> '+' '+' <EXPR>,
        return ctuple(C_IMPPLUSPLUS, -1, compile_expression(child(astp,4), 0, VALUE_WANTED));
      } else if (alt == 2) {     //\\   <immediate:return !in_const_initialiser_list;> '-' '-' <EXPR>,
        return ctuple(C_IMPMINUSMINUS, -1, compile_expression(child(astp,4), 0, VALUE_WANTED));
      } else if (alt == 3) {     //\\   <OPT-ACTUAL-PARAMETERS>;
        return compile_expression(child(astp,1), 0, VALUE_WANTED); // C_PARAMETER_LIST or -1
      }

    //\\ #  Procedure application, i.e. actual parameters to routine or function call
    case AST_OPT_ACTUAL_PARAMETERS://\\ P<OPT-ACTUAL-PARAMETERS> =
      if (alt == 0) {            //\\   <immediate:return !in_const_initialiser_list;> '(' <EXPR> <REST-OF-ACTUAL-PARAMETERS> ')',
        int expr, rest;
        // COULD BE '[' IF ARRAY RATHER THAN PROC. TO DO. We'll get to that later...
        expr = compile_expression(child(astp,3), 0, VALUE_WANTED);
        rest = compile_expression(child(astp,4), 0, VALUE_WANTED);
        //cps(")");
        return ctuple(C_BRACKETED_LIST, -1, expr, rest); // #1 object this hangs off  #2 actual parameter  #3 rest of actual parameter list (another C_PARAMETER_LIST)
      } else if (alt == 1) {     //\\   ;
        return -1; // no parameters
      }

    case AST_Opt_SUBFIELD:       //\\ P<Opt-SUBFIELD> =
                                 //\\   <immediate:return !in_const_initialiser_list;> '_' <SUBFIELD>,
                                 //\\   ;
    case AST_Opt_RECORDFIELD:    //\\ P<Opt-RECORDFIELD> =
      if (alt == 0) {            //\\   <immediate:return !in_const_initialiser_list;> '_' <RECORDFIELD>,
        //                                                                                 ^
        // <lvalue> at this point could maybe be a new grammar type - it's not really an lvalue and
        // to build the CAST properly we need to pass information down from the top rather than building
        // up from below.  Might be able to make this work without too much redesign if I passed the
        // uncompiled AST back up rather than the compiled CAST and postponed the calls to compile()
        // until later.

        return ctuple(C_RECORD_FIELD,  // need to make '_' in parent be followed by <LVALUE>
                      -1 /* to be filled in by parent*/,
                      compile_expression(child(astp,3) /* the lvalue. At this stage we're just recording the name and mirroring the AST.
                                                          Declaration/recordfield checking and CAST comes later because the children
                                                          on the RHS require type information from the parents to be passed down. */,
                                         0/*(for the missing type info)*/,
                                         VALUE_WANTED),
                      -1 /* is pointer? */); // don't know at this point of parent is record or recordname
      } else if (alt == 1) {     //\\   ;
        return -1;
      }

    case AST_SCALAR_ASSIGN:      //\\ P<SCALAR-ASSIGN> =
    {
      if (alt == 0) {            //\\   '=',
        return ctuple(C_ASSIGN_VALUE, -1, -1);
      } else if (alt == 1) {     //\\   '<' '-';
        return ctuple(C_JAM_TRANSFER, -1, -1);
      }
    }

    //\\ #  Need <EXPR> instead of <CEXPR> to catch = tostring(i) because cexpr only got tostring
    //\\ #  Need '== <EXPR>' instead of '== <LVALUE>' to catch '== length(s)', i.e. a %map although it looks like a fn call.
    //\\ #  - the latter being more of a failing of LVALUE that here.
    //\\ #  Removed:
    //\\ #    '==' <LVALUE>,
    //\\ #    '=' <CEXPR>,
    //\\ #  These will be improved once we have semantic checks...
    //\\ # 

    //\\ # ONLY used now for initialising arrays. Should move to <Opt-Init-assign-array>
    //    **** NOT MAYBE TRUE.  Is it being used for for-loop control?
    // Idiot.  It is also used in %result == lvalue ...
    case AST_ASSOP:
    {
      if (alt == 0) {            //\\   '=' '=',
        return ctuple(C_ASSIGN_ADDRESS, -1, -1);
      } else if (alt == 1) {     //\\   <SCALAR-ASSIGN>;
        int assign = compile_expression(leftchild(astp), 0, VALUE_WANTED); // SCALAR-ASSIGN now returns a ctuple
        // P<SCALAR-ASSIGN> = '=' | '<' '-' ;
        // now returns a ctuple of C_ASSIGN_VALUE or C_JAM_TRANSFER
        if (AstOP(assign) == C_ASSIGN_VALUE) {
          return assign;
        } else if (AstOP(assign) == C_JAM_TRANSFER) {
          return assign;
        } else { //BUG
          diagnose(assign);
          fault("Internal error - unexpected C AST type %s at line %d", CAstOPName(AstOP(assign)), __LINE__);
        }
      }
    }

    //\\ P<ASSOP-EXPR> =
    //\\    '=' '=' <LVALUE>,
    //\\    <SCALAR-ASSIGN> <EXPR>;
    
    case AST_ASSOP_EXPR:
      if (alt == 0) {            //   '=' '=' <LVALUE>,
        // much simplified. Except where it isn't.
        int result = compile_expression(child(astp, 3), 0, VALUE_WANTED);
        result = get_address_of(result);
        return ctuple(C_ASSIGN_ADDRESS, -1, result);

      } else if (alt == 1) {     //   <SCALAR-ASSIGN> <EXPR>;
        int assign_c = compile_expression(leftchild(astp), 0, VALUE_WANTED); // C_ASSIGN_VALUE or C_JAM_TRANSFER
        int expr_c   = compile_expression(rightchild(astp), 0, VALUE_WANTED);
        assert(rightchild(assign_c) == -1);
        rightchild(assign_c) = expr_c;
        return assign_c;

      }

      
    //\\ #  handling precedence with the grammar rather than Edinburgh-style in the AST. Tradeoff of speed for simplicity.
    //\\ #  A <CEXPR> does not parse differently from an <EXPR> - it just checks after building the tree that all the
    //\\ #  elements were constant and can be folded.

    case AST_CEXPR: //\\ P<CEXPR> = <guard...> <EXPR> <cancel guard...> ;
    {
      int expr     = compile_expression(child(astp,2), 0, VALUE_WANTED);
      int constval = IntConstant_c(expr);
      if (constval == -1) {
        if (AstOP(expr) != C_CONST_STRING) {
          char source[1024];
          getstr(child(astp,2), source);
          warn("%s is not an integer constant", source); diagnose(astp);
        }
        return expr;
      }
      return constval;
    }
    
      // expr = 
      //              <rest-of-expr#1>
      //                 /
      //                /  
      //         <mediumprec-expr#1>

      // plus

      // rest-of-expr#1 =
      //
      //                <rest-of-expr#2>
      //                       /
      //                      /
      //                  <op-low>
      //                    /  \
      //                   /    \
      //                 -1      <mediumprec-expr#2>
      
      // becomes
      
      // expr =
      //                <op-low>
      //                  /  \
      //                 /    \
      // <mediumprec-expr#1>   <mediumprec-expr#2>
      //


      //\\ P<EXPR> =
      //\\    <DOTTED-STRING-EXPR>,
      //\\    <MEDIUMPREC-EXPR> <REST-OF-EXPR>;

    case AST_EXPR:
    {
      if (alt == 0) return compile_expression(leftchild(astp), 0, VALUE_WANTED); // <DOTTED-STRING-EXPR>, // because "." operator removed from regular exprs!
      int left_operand = compile_expression(leftchild(astp), 0, VALUE_WANTED);   // <MEDIUMPREC-EXPR>
      int top_op       = compile_expression(rightchild(astp), 0, VALUE_WANTED);  // <REST-OF-EXPR>
      int left_op      = top_op;
      if (top_op == -1) return left_operand;
      while (leftchild(left_op) != -1) left_op = leftchild(left_op); leftchild(left_op) = left_operand;
      return top_op;
    }

    case AST_REST_OF_EXPR:
    {
      if (alt == 1) return -1;
      int left_op       = compile_expression(leftchild(astp), 0, VALUE_WANTED); // <OP-LOW> an empty ctuple(-1,-1)
      int right_operand = compile_expression(rightchild(astp), 0, VALUE_WANTED); // <MEDIUMPREC-EXPR>
      int top_op        = compile_expression(child(astp, 3), 0, VALUE_WANTED); // <REST-OF-EXPR>
      int hole_op       = top_op;
      rightchild(left_op) = right_operand;
      if (top_op == -1) return left_op;
      while (leftchild(hole_op) != -1) hole_op = leftchild(hole_op); leftchild(hole_op) = left_op;
      return top_op;
    }

    case AST_REAL_EXP:           //\\ P<REAL-EXP> =
      if (alt == 0) {            //\\   '*' '*',
        return ctuple(C_REXP, -1, -1);
      } else if (alt == 1) {     //\\   '\' <!eq>,
        return ctuple(C_REXP, -1, -1);
      } else if (alt == 2) {     //\\   '^';
        return ctuple(C_REXP, -1, -1);
      }
      
    case AST_INTEGER_EXP:   //\\   '*' '*' '*' '*', '\' '\',  '^' '^';
    return ctuple(C_IEXP, -1, -1);

    //\\ #  highest precedence, right-associative (only exponentiation)
    //\\ # 
    //\\ #  I found this expression the 68000 Imp compiler:
    //\\ # 
    //\\ #            y = |r|+0.5/10.0^m;  !modulus, rounded
    //\\ # 
    //\\ #  which I parse as:
    //\\ # 
    //\\ #            y = |r|+( 0.5 / (10.0^m));  !modulus, rounded
    //\\ # 
    //\\ #  I presume this is a clever way to round the last digit of a decimal,
    //\\ #  but damn, how it works is beyond me.
    //\\ #  Typical Hamish code! Brilliant (I presume) but obscure and uncommented.
    //\\ # 
    case AST_OP_HIGH_RIGHTASSOC:
    if (alt == 0) {            //\\   <INTEGER-EXP>,
      return compile_expression(leftchild(astp), 0, VALUE_WANTED);
    } else if (alt == 1) {     //\\   <REAL-EXP>;
      return compile_expression(leftchild(astp), 0, VALUE_WANTED);
    }

    case AST_RIGHTASSOC_EXPR_with_OP:
    if (alt == 0) {            //\\   <UNARY-EXPR> <OP-HIGH-RIGHTASSOC>;
      int left_operand = compile_expression(leftchild(astp), 0, VALUE_WANTED);
      int op           = compile_expression(rightchild(astp), 0, VALUE_WANTED);
      leftchild(op) = left_operand;
      return op;
    }

    //\\ #  <OP> is only used in <CEXPR>s now and will be removed entirely
    //\\ #  once CEXPRs are recoded to use the same grammar as EXPRs.
    //\\ # 
    //\\ #  Note that none of these options eliminates the possibility of implementing
    //\\ #  conditional compilation the EMAS Imp80 way using
    //\\ # 
    //\\ #  %if <CEXPR> %start
    //\\ #    ! conditional code
    //\\ #  %finish
    //\\ # 
    //\\ #  (Though I've always had questions about that on EMAS if there's a
    //\\ #   label inside the block - the ability to use that construct for
    //\\ #   conditional compilation depends on whether that label is ever
    //\\ #   jumped to from code outside the block! - doing it properly requires
    //\\ #   full flow control analysis and a transitive closure on the flow graph!)
    //\\ # 

  //\\ #  highest precedence, left-associative  
    case AST_OP_HIGH_LEFTASSOC:
    if (alt == 0) {            //\\   '<' '<',
      return ctuple(C_SHL, -1, -1);
    } else if (alt == 1) {     //\\   '>' '>';
      return ctuple(C_SHR, -1, -1);
    }

  //\\ #  middle-precedence operators.  All left-associative.
    case AST_OP_MED:             //\\ P<OP-MED> =
    if (alt == 0) {            //\\   '*',
      return ctuple(C_MUL, -1, -1);
    } else if (alt == 1) {     //\\   '/' '/',
      return ctuple(C_IDIV, -1, -1);
    } else if (alt == 2) {     //\\   '/',
      return ctuple(C_RDIV, -1, -1);
    } else if (alt == 3) {     //\\   '&';
      return ctuple(C_BITAND, -1, -1);
    }

  //\\ #  fortunately unary backslash is unambiguous with ERCC style exponentiation binary operator
    case AST_UNARY_NOT:          //\\ P<UNARY-NOT> =
      if (alt == 0) {            //\\   '\',
        return ctuple(C_BITNOT, -1);
      } else if (alt == 1) {     //\\   '~';
        return ctuple(C_BITNOT, -1);
      }

  //\\ #  lowest precedence operators.  All left-associative.
    case AST_OP_LOW:             //\\ P<OP-LOW> =
    if (alt == 0) {            //   '+',
      return ctuple(C_BINPLUS, -1, -1);
    } else if (alt == 1) {     //   '-',
      return ctuple(C_BINMINUS, -1, -1);
    } else if (alt == 2) {     //   '!' '!',
      return ctuple(C_BITEXOR, -1, -1);
    } else if (alt == 3) {     //   '!';
      return ctuple(C_BITOR, -1, -1);
    }
    return TO_DO("OP_LOW dropping through by accident?");
    // '.' removed.

    case AST_MEDIUMPREC_EXPR:                         //\\ P<MEDIUMPREC-EXPR> =
                                                    //\\    <HIGHPREC-EXPR> <REST-OF-MEDIUMPREC-EXPR>;
    {
      int left_operand = compile_expression(leftchild(astp), 0, VALUE_WANTED);  // <HIGHPREC-EXPR>
      int top_op       = compile_expression(rightchild(astp), 0, VALUE_WANTED); // <REST-OF-MEDIUMPREC-EXPR>
      if (top_op == -1) return left_operand;
      int hole_op = top_op;
      while (leftchild(hole_op) != -1) hole_op = leftchild(hole_op); leftchild(hole_op) = left_operand;
      return top_op;
    }

    case AST_REST_OF_MEDIUMPREC_EXPR:                 //\\ P<REST-OF-MEDIUMPREC-EXPR> =
                                                    //\\    <OP-MED> <HIGHPREC-EXPR> <REST-OF-MEDIUMPREC-EXPR>,
                                                    //\\   ;
    {
      if (alt == 1) return -1;
      int left_op       = compile_expression(leftchild(astp), 0, VALUE_WANTED);  // <OP-MED> - returns an empty C_op tuple, with left and right both -1
      int right_operand = compile_expression(rightchild(astp), 0, VALUE_WANTED); // <HIGHPREC-EXPR>
      int top_op        = compile_expression(child(astp, 3), 0, VALUE_WANTED);   // <REST-OF-MEDIUMPREC-EXPR>
      rightchild(left_op) = right_operand; // leftchild already -1
      if (top_op == -1) return left_op;    // left operand will be filled in by parents
      int hole_op = top_op;
      while (leftchild(hole_op) != -1) hole_op = leftchild(hole_op); leftchild(hole_op) = left_op;
      return top_op;
    }
    
    case AST_HIGHPREC_EXPR:
    {                                                 // P<HIGHPREC-EXPR> = <RIGHT-ASSOC-EXPR> <REST-OF-HIGHPREC-EXPR>;
      int left_operand = compile_expression(leftchild(astp), 0, VALUE_WANTED);    // <RIGHT-ASSOC-EXPR>
      int top_op       = compile_expression(rightchild(astp), 0, VALUE_WANTED);   // <REST-OF-HIGHPREC-EXPR>
      if (top_op == -1) return left_operand;
      int hole_op = top_op;
      while (leftchild(hole_op) != -1) hole_op = leftchild(hole_op); leftchild(hole_op) = left_operand;
      return top_op;
    }

    case AST_REST_OF_HIGHPREC_EXPR:
    {
      if (alt == 1) return -1;
      int left_op       = compile_expression(leftchild(astp), 0, VALUE_WANTED);  // <OP-HIGH-LEFTASSOC>
      int right_operand = compile_expression(rightchild(astp), 0, VALUE_WANTED); // <RIGHT-ASSOC-EXPR>
      int top_op        = compile_expression(child(astp, 3), 0, VALUE_WANTED);   // <REST-OF-HIGHPREC-EXPR>
      rightchild(left_op) = right_operand;
      if (top_op == -1) return left_op;
      int hole_op = top_op;
      while (leftchild(hole_op) != -1) hole_op = leftchild(hole_op); leftchild(hole_op) = left_op;
      return top_op; 
    }

    case AST_RIGHT_ASSOC_EXPR:
    {  
      int top_op            = compile_expression(leftchild(astp), 0, VALUE_WANTED);  // <REST-OF-RIGHTASSOC-EXPR>
      int rightmost_operand = compile_expression(rightchild(astp), 0, VALUE_WANTED); // <UNARY-EXPR>
      if (top_op == -1) return rightmost_operand;        // <UNARY-EXPR> only, nothing on the left of expr.
      int hole_op = top_op;
      while (rightchild(hole_op) != -1) hole_op = rightchild(hole_op); rightchild(hole_op) = rightmost_operand;
      return top_op;
    }

    case AST_REST_OF_RIGHTASSOC_EXPR:
    {
      if (alt == 1) return -1;
      int top_op        = compile_expression(leftchild(astp), 0, VALUE_WANTED);      // <REST-OF-RIGHTASSOC-EXPR>
      int right_operand = compile_expression(rightchild(astp), 0, VALUE_WANTED);
      if (right_operand == -1) return top_op; // nothing on the left of expr.
      rightchild(top_op) = right_operand;
      return top_op;
    }
    
    case AST_UNARY_EXPR:
    {
      int op      = compile_expression(leftchild(astp), 0, VALUE_WANTED); // <Opt-UNARY-OP>
      int operand = compile_expression(rightchild(astp), 0, VALUE_WANTED); // <OPERAND>
      if (op == -1) return operand;
      leftchild(op) = operand;
      return op;
    }

    //\\ #  You would think that unary operators always have the highest precedence,
    //\\ #  but not so - here's a tricky one...
    //\\ # 
    //\\ #  I = -1 >> 1
    //\\ # 
    //\\ #  This always has the value 0, regardless of integer size, because a unary minus
    //\\ #  in front of a constant is *NOT* interpreted as a negative constant, but rather
    //\\ #  the positive constant (1 in this example), preceded by a unary negation operator.
    //\\ # 
    //\\ #  It gets worse... unary negation is implemented as binary subtraction with
    //\\ #  an inserted '0' left operand, i.e.
    //\\ # 
    //\\ #  0 - 1 >> 1
    //\\ # 
    //\\ #  This appears to have been specified at the lexical level and no implicit bracketing is mentioned
    //\\ #  Shifts have higher precedence than subtraction, so this becomes:
    //\\ # 
    //\\ #  0 - (1 >> 1)
    //\\ # 
    //\\ #  which evaluates to 0.
    //\\ # 
    //\\ #  The 68000 Imp compiler includes this statement which I hope is correct! dict(firstpos)_val = -cad>>1
    //\\ # 
    //\\ #  Some versions of Imp (I think Imp80) forbid a unary '+' before constants,
    //\\ #  so A + +3 would be illegal (although A + -3 would not)
    //\\ # 
    //\\ #  and none of this applies to binary-not (\) so \3 really is higher
    //\\ #  priority than '>>' and \3 >> 1 does what you would expect.  I hope.
    //\\ # 
    case AST_Opt_UNARY_OP:
    {
      if (alt == 0) {            //\\   '+',
        return ctuple(C_UNARY_PLUS, -1);
      } else if (alt == 1) {     //\\   '-',
        return ctuple(C_UNARY_MINUS, -1);
      } else if (alt == 2) {     //\\   <UNARY-NOT>,
        return compile_expression(leftchild(astp), 0, VALUE_WANTED);
      } else if (alt == 3) {     //\\   ;
        return -1;
      }
    }

    //\\ #  The old form of modulus (!X!) is for really ancient compilers such as Hamish's Pdp15 Imp.
    //\\ #  Unfortunately there are obscure cases where an ambiguous parse results:
    //\\ # 
    //\\ #   putact(condop!!polarity!!1+polarity<<7,item,0)
    //\\ # 
    //\\ #  Does that mean
    //\\ #    putact(condop !!  polarity   !!  1 + polarity<<7, item, 0)
    //\\ #  or
    //\\ #    putact(condop !  !polarity!  !   1 + polarity<<7, item, 0)
    //\\ # 
    //\\ #  I need to check but I *hope* that the grammar here causes the first of the above to be selected.
    //\\ #  If not, we'll need a command-line option here to force the version of Imp being used.
    //\\ # 
    //\\ #  I just hope that there's no version of Imp that ever supported factorials with a postfix '!'
    //\\ # 
    case AST_OPERAND:                 //\\ P<OPERAND> =
    {
      if (alt == 0) {               //\\    <REAL-CONST>,
        return compile_expression(leftchild(astp), 0, VALUE_WANTED);
      } else if (alt == 1) {        //\\    <INT-CONST>,
        return compile_expression(leftchild(astp), 0, VALUE_WANTED);
      } else if (alt == 2) {        //\\    <STR-CONST>,
        return compile_expression(leftchild(astp), 0, VALUE_WANTED);
      } else if (alt == 3) {        //\\    <LVALUE>,
        return compile_expression(leftchild(astp), 0, VALUE_WANTED);
      } else if (alt == 4) {     //\\   '(' <EXPR> ')',
        return ctuple(C_BRACKET, compile_expression(rightchild(astp),0,VALUE_WANTED), 0, VALUE_WANTED);
      } else if (alt == 5) {     //\\   '|' <EXPR> '|',
        return ctuple(C_ABS, compile_expression(rightchild(astp),0,VALUE_WANTED), 0, VALUE_WANTED);
      } else if (alt == 6) {     //\\   '!' <EXPR> '!';
        return ctuple(C_ABS, compile_expression(rightchild(astp),0,VALUE_WANTED), 0, VALUE_WANTED);
      }
    }

    //\\ # Roughly... 
    //\\ # # P<basedconst> =  "[0-9][0-9]*_[0-9A-Z][0-9A-Z]*"
    //\\ # btw regexps are a planned future enhancement to the parser.
    case AST_INT_CONST:          //\\ P<INT-CONST> = <SAVED-INT-CONST>;
    return compile_expression(leftchild(astp), 0, VALUE_WANTED);

    case AST_SAVED_INT_CONST:    //\\ P<SAVED-INT-CONST> =
    {
      int i;
      if (alt == 0) {            //\\   <DIGIT-SEQ> '_' <alphanumeric-SEQ>,
        // TO DO: return C_ objects for all consts.
        int base_c = compile_expression(leftchild(astp), 0, VALUE_WANTED);
        int val_c  = compile_expression(child(astp, 3), 0, VALUE_WANTED);
        int rc, base, value;
        char num[68];
        // TO DO: digit seqs - remove leading zeroes...
        rc = sscanf(pool_to_str(leftchild(base_c)), "%d", &base);
        if (base == 2) {
          sprintf(num, "0b%s", pool_to_str(leftchild(val_c)));
          return ctuple(C_CONST_INT, str_to_pool(num), binstr_to_int(pool_to_str(leftchild(val_c))));
        } else if (base == 8) {
          sprintf(num, "0%s", pool_to_str(leftchild(val_c)));
          return ctuple(C_CONST_INT, str_to_pool(num), octstr_to_int(pool_to_str(leftchild(val_c))));
        } else if (base == 10) {
          sprintf(num, "%s", pool_to_str(leftchild(val_c)));
          return ctuple(C_CONST_INT, str_to_pool(num), str_to_int(pool_to_str(leftchild(val_c))));
        } else if (base == 16) {
          char *s = pool_to_str(leftchild(val_c));
          value = 0;
          while (*s != '\0') {
            int c = (*s++)&255;
            if ('0' <= c && c <= '9') {
              value = (value << 4) | (c-'0');
            } else if ('a' <= c && c <= 'f') {
              value = (value << 4) | (c-'a'+10);
            } else if ('A' <= c && c <= 'F') {
              value = (value << 4) | (c-'A'+10);
            } else {
              fault("Bad HEX constant: %s_%s", leftchild(base_c), leftchild(val_c));
            }
          }
          sprintf(num, "0x%s", pool_to_str(leftchild(val_c)));
          //fprintf(stderr, "*** Creating C_CONST_INT 16_ <%s,%d>\n", num, value);
          return ctuple(C_CONST_INT, str_to_pool(num), value);
        } else {
          warn("Unimplemented base conversion %0d", base);
          return ctuple(C_CONST_INT, str_to_pool("BAD BASE"), 999888);
        }
      } else if (alt == 1) {     //\\   <DIGIT-SEQ>,
        int digits = compile_expression(leftchild(astp), 0, VALUE_WANTED);
        int value, rc;
        rc = sscanf(pool_to_str(leftchild(digits)), "%d", &value);
        if ((rc != 1) || (errno != 0)) {
          long long bigint;
          errno = 0;
          rc = sscanf(pool_to_str(leftchild(digits)), "%lld", &bigint);
          if ((rc != 1) || (errno != 0)) {
            fault("Invalid integer %s - %s", pool_to_str(leftchild(digits)), strerror(errno));
          } else {
            // Older IMP compilers would have converted this to a double (%longreal)
            // and passed it back up as a real constant.  But we'll support %longinteger like EMAS3...
            return ctuple(C_CONST_BIGINT, leftchild(digits), (int)((bigint>>32LL)&0xFFFFFFFFLL), (int)(bigint&0xFFFFFFFFLL));
          }
        }
        return ctuple(C_CONST_INT, leftchild(digits), value);
      } else if (alt == 2) {     //\\   <OLDSTYLE-BASE>,
        return compile_expression(leftchild(astp), 0, VALUE_WANTED);

      } else if (alt == 3) {     //\\   <sqstring>; // BIP
        char *s;
        int value;
        int sqstring  = compile_expression(leftchild(astp), 0, VALUE_WANTED); // saved in canonical form

        // fprintf(stderr, "INFO: type of sqstring is %s\n", CAstOPName(AstOP(sqstring)));  - C_STRING_LITERAL
        // if (strlen(str_lit) == 0 || strlen(str_lit) > 4) fault("Bad quoted constant: '%s'", str_lit);
        // more likely old-style string, in which case this would not be the place to complain.
        // However if that is the case, we should return a C_CONST_STRING instead...
        s = STRING_LITERAL_to_C_int_const(sqstring, &value); // remove doubled quotes ('') from sqstring. Enclose string in single quotes.
        //fprintf(stderr, "STRING_LITERAL_to_C_int_const: %s (%d)\n", s, value);
        return ctuple(C_CONST_INT, str_to_pool(s), value); // value should be valid if form was 'x' or 'xxxx' or M'xxxx' ...

      }
    }
    
    default:
    fprintf(stderr, "MISSING AST OPERATOR AST_%s IN compile_expression(@AST:%d, %s)\n",
            AstOPName(astop),
            astp, (address_wanted ? "GET ADDRESS" : "GET VALUE"));

    // pending relocation from compile()?
    case AST_REST_OF_ACTUAL_PARAMETERS:
    case AST_STRING_RESOLUTION_EXPR:    
    // These ones we know we have to call back for:
    case AST_NAME:
    case AST_DIGIT_SEQ:
    case AST_REAL_CONST:
    case AST_STR_CONST:
    case AST_alphanumeric_SEQ:
    case AST_DOTTED_STRING_EXPR:
    case AST_OLDSTYLE_BASE:
    case AST_sqstring:
    return compile_inner(astp, extra_parameter, line);
  }
}

int generate_expression(int astp, int address_wanted) {
  int castp;
  if (debug_new_exprs) fprintf(stderr, "generate_expression(@AST:%d, %s)\n", astp, (address_wanted ? "GET ADDRESS" : "GET VALUE"));
  castp = compile_expression(astp, 0, address_wanted);
  output_c_expression(castp);
}
//0x0041e9b8 in compile_expression_inner (astp=89345, extra_parameter=-1, address_wanted=0, line=3503) at ./expressions.c:255
//255               leftchild(assign_c) = compile_expression(child(astp,1), 0, VALUE_WANTED);
