// DO NOT EDIT THIS FILE.
// This code has been re-inserted into compile.c (from where it was originally extracted) to
// make it easier to edit the code generation without switching back and forth between files.

/*
    replacement code and data structures for Imp expression evaluation...

    ctuples for an operator are a pair of <C_ and descriptor> - the descriptor
    may be a declaration symbol but it could also be the same information for
    an expression, constructed bottom-up, to allow type-checking at the top,
    or appropriate casting for assignments/jam transfers etc.
 */

#define CAstOP(x) AstOP(x)

int debug_flowcontrol = 1;

#ifdef NEVER
char *CAstOPName(int castop) {
  return CAST[castop-C_BASE].name;
}

char *AstOPName(int astop) {
  return phrasename[astop];
}
#endif

void output_c_flowcontrol(int castp) {
  if (debug_flowcontrol) fprintf(stderr, "output_c_flowcontrol: @C_AST:%d\n", castp);
  if (castp < 0) {
    return;
  }
  int castop = CAstOP(castp);
  switch (castop) {


    
  default:
    fprintf(stderr, "MISSING CAST OPERATOR C_%s IN output_c_flowcontrol: @C_AST:%d\n",
            CAstOPName(castop),
            castp);
    return;
  }
}

#define compile_flowcontrol(astp, extra_parameter) compile_flowcontrol_inner(astp, extra_parameter, __LINE__)
int compile_flowcontrol_inner(int astp, int extra_parameter, int line) {
  //if (debug_flowcontrol) fprintf(stderr, "compile_flowcontrol(@AST:%d, %d) from line %d\n", astp, extra_parameter, line);
  if (astp < 0) {
    fprintf(stderr, "compile_flowcontrol: -1 from line %d\n", line);
    return -1;
  }
  int astop = AstOP(astp);
  int alt      = Alt0(astp);
  int count    = AST[astp+AST_count_offset];
  int *A_child = &AST[astp+AST_child1_offset]; // An index into a child of the AREC, *not* a child of the AST.
                                               // (for printing terminals with RECONSTITUTE, although it may be better
                                               //  to pass astp and let RECONSTITUTE apply AST_child1_offset)
  if (astop >= C_BASE) {
    fprintf(stderr, "? redundant compile(C_%s) called from line %d\n", CAstOPName(astop), line);
    return astp; // already compiled.
  }


  switch (astop) {

  case AST_TOP_LEVEL_CONDITION: //\\ P<TOP-LEVEL-CONDITION> = <CONDITION>;
      return compile(leftchild(astp));

  case AST_Opt_UNTIL:          //\\ P<Opt-UNTIL> =
      if (alt == 0) {            //\\   "until" <TOP-LEVEL-CONDITION>,
        return ctuple(C_UNTIL, -1, compile(rightchild(astp)));
      } else if (alt == 1) {     //\\   ;
        return -1;
      }

  case AST_Percent_IU:         //\\ P<Percent-IU> =
      if (alt == 0) {            //\\   "if",
        return ctuple(C_IF, -1, -1, -1, -1);            // #1 'then' statement to be executed  #2 cond  #3 'else' statement to be executed
      } else if (alt == 1) {     //\\   "unless";
        return ctuple(C_UNLESS, -1, -1, -1, -1);        // #1 'then' statement to be executed  #2 cond  #3 'else' statement to be executed
      }

    //\\ #  Added old-style while/until for: %UNTIL NEXTINSYM=NL %THEN %CYCLE
    //\\ #  and %WHILE NEXTINSYM#NL %THEN SKIPINSYM
    case AST_CYCLE_S:            //\\ P<CYCLE-S> =
      if (alt == 0) {            //\\   "cycle" <Opt-CYCPARM> <S> <INTERNALS> "repeat" <Opt-UNTIL> <S>,
        int cycle = compile(child(astp,2)); // Either I=1,2,3 style or -1 (plain %cycle)
        int body  = compile(child(astp,4));
        int until = compile(child(astp,6)); // Either "until <COND>" (C_UNTIL) or -1.

        if (cycle == -1) { // starts with a plain %cycle
          if (until != -1) { // ends in a %until <cond>
            child(until, 1) = body;
            return until;
          } else {  // ends in a plain %repeat
            return ctuple(C_FOREVER, body);
          }
        }

        assert(AstOP(cycle) == C_IMP_FOR); // old syntax for-loop.
        child(cycle, 1) = body;
        return cycle;

      } else if (alt == 1) {     //\\   <Percent-WUF> "cycle" <S> <INTERNALS> "repeat" <Opt-UNTIL> <S>,
        int wuf = compile(child(astp,1)); // while/until/for - no null option
        int block = compile(child(astp,4));
        int opt_until = compile(child(astp,6));
        if (opt_until != -1) {
          fault("A %%cycle starting with %%while, %%until, or %%for must not end with %%repeat %%until.");
        }
        leftchild(wuf) = STATEMENT_BLOCK(block);
        return wuf;

      } else if (alt == 2) {     //\\   <Percent-WU> "then" "cycle" <S> <INTERNALS> "repeat" <Opt-UNTIL> <S>,
        int wu        = compile(child(astp,1)); // while/until - no null option
        int block     = compile(child(astp,5));
        int opt_until = compile(child(astp,7));
        assert(AstOP(wu) == C_WHILE || AstOP(wu) == C_UNTIL);
        assert(leftchild(wu) == -1);
        if (opt_until != -1) {
          fault("A %%cycle starting with %%while, %%until, or %%for must not end with %%repeat %%until.");
        }
        if (block != -1 && AstOP(block) != C_SEQ) block = ctuple(C_SEQ, block, -1);
        leftchild(wu) = STATEMENT_BLOCK(block);
        return wu;

      } else if (alt == 3) {     //\\   <Percent-WU> "then" <UI>;
        int wu        = compile(child(astp,1)); // while/until - no null option
        int block     = STATEMENT_BLOCK(compile(child(astp,3)));
        assert(AstOP(wu) == C_WHILE || AstOP(wu) == C_UNTIL);
        assert(leftchild(wu) == -1);
        block = ctuple(C_SEQ, block, -1);
        leftchild(wu) = block;
        return wu;

      }

    case AST_IF_S:               //\\ P<IF-S> =
      if (alt == 0) {            //\\   <Percent-IU> <TOP-LEVEL-CONDITION> <THEN-S>;

        // We note the condition as an extra field after re-evaluating it as a constant, so that
        // we can do conditiona compilation the EMAS way.  HOWEVER that should only be done if
        // a compile-time flag allows it, *and* when we're not allowing it, we need to check
        // the scope level so that we don't allow %if statements outside a block.

        // *** NOTE *** The current implementation does not handle declarations made within the
        // then/else blocks of conditional start/finish statements.  Need to work out a way to
        // only execute declarations for the TRUE arm. Unfortunately we are handling declarations
        // at compile() time, not out() time.  So some restructuring definitely needed...

        // The snswer is going to be related to calling compile() of the then/else blocks only
        // if the constant condition was true.
        
        int if_or_unless = compile(child(astp,1));        assert(AstOP(if_or_unless) == C_IF || AstOP(if_or_unless) == C_UNLESS);
        int cond         = compile(child(astp,2));        assert(cond != -1);
        int then_else    = compile(child(astp,3));        assert(child(if_or_unless, 1) == -1);        assert(child(if_or_unless, 2) == -1);
        int then, else_c;
        if (AstOP(then_else) == C_THEN_ELSE) {
          detuple(then_else, C_THEN_ELSE, &then, &else_c);
        } else {
          then   = child(if_or_unless, 1);
          else_c = child(if_or_unless, 3);
        }
        if (AstOP(if_or_unless) == C_UNLESS) {
          cond = ctuple(C_LOGNOT, cond);
          AstOP(if_or_unless) = C_IF;
        }
        assert(AstOP(if_or_unless) == C_IF);
        child(if_or_unless, 2) = cond;
        child(if_or_unless, 4) = IntConstant_c(cond);
        child(if_or_unless, 1) = STATEMENT_BLOCK(then);
        if (else_c != -1) else_c = STATEMENT_BLOCK(else_c);
        child(if_or_unless, 3) = else_c;
        return if_or_unless;

      }
      break;

    //\\ #  /home/gtoal/gtoal.com/athome/edinburgh/APM-gdmr/I/LIB.imp
    //\\ #  contains:
    //\\ #           %elseif 'A'<=sym<='Z' %start
    //\\ # 
    //\\ #  which is a form of Hamish's that is equivalent to
    //\\ # 
    //\\ #   %finish %else %if 'A'<=sym<='Z' %start
    //\\ # 
    //\\ #  or the more concise
    //\\ # 
    //\\ #   %else %if 'A'<=sym<='Z'
    //\\ # 
    //\\ #  An identical issue is: %elsestart
    //\\ #  in APM-gdmr/ETHER/PROFILE.imp
    //\\ # 


    // it is either finishing a start or an elseif...start ...
    case AST_FINISH_S:           //\\ P<FINISH-S> =
      if (alt == 0) {            //\\   "else" <Percent-IU> <TOP-LEVEL-CONDITION> <Opt-start> <S> <INTERNALS> <FINISH-S>,
                                 //            \_____________________________________________________________/
                                 //                                         |
                                 //                                     statement

        int then_block  = STATEMENT_BLOCK(compile(child(astp,6)));
        int finish      = compile(child(astp,7));
        int if_or_unless = compile(child(astp,2));        assert(AstOP(if_or_unless) == C_IF || AstOP(if_or_unless) == C_UNLESS);
        int cond         = compile(child(astp,3));        assert(cond != -1);
        int statement = if_or_unless;

        if (AstOP(if_or_unless) == C_UNLESS) { // convert %unless to %if %not ...
          cond = ctuple(C_LOGNOT, cond);
          AstOP(if_or_unless) = C_IF;
        }
        assert(AstOP(if_or_unless) == C_IF);
        
        child(if_or_unless, 1) = then_block;
        child(if_or_unless, 2) = cond;
        child(if_or_unless, 4) = IntConstant_c(cond);
        if (finish != -1) finish = STATEMENT_BLOCK(finish);
        child(if_or_unless, 3) = finish; // the else part
        
        if (finish != -1 && AstOP(finish) == C_THEN_ELSE) {          assert(leftchild(finish) == -1);
          assert(leftchild(finish) == -1);
          int else_c = rightchild(finish);
          if (else_c != -1) else_c = STATEMENT_BLOCK(else_c);
          child(if_or_unless, 3) = else_c;
        }

        return ctuple(C_THEN_ELSE, -1, statement);

      } else if (alt == 1) {     //\\   "else" <Opt-start> <S> <INTERNALS> "finish" <S>,
        int block = STATEMENT_BLOCK(compile(child(astp,4)));
        return ctuple(C_THEN_ELSE, -1, block);

      } else if (alt == 2) {     //\\   "finish" <Opt-ELSE-S>;
        int opt_else = compile(rightchild(astp));
        if (opt_else == -1) return -1; // plain finish returns -1
        assert(AstOP(opt_else) == C_THEN_ELSE);
        return opt_else;

      }

    case AST_Opt_start:          //\\ P<Opt-start> = "start", ;
      // cps("{");
      return -1;  // syntactic sugar

    //\\ #  see the comments on <FINISH-S> above.
    case AST_Opt_ELSE_S:         //\\ P<Opt-ELSE-S> =
      // returns a C_ELSE or -1
      if (alt == 0) {            //\\   "else" "start" <S> <INTERNALS> "finish" <S>,
        int block = STATEMENT_BLOCK(compile(child(astp,4)));
        return ctuple(C_THEN_ELSE, -1, block);

      } else if (alt == 1) {     //\\   "else" <IF-S>,
        // may decide to add an 'else if' to preserve indentation.  Meanwhile...
        int block  = compile(child(astp,2));
        int else_c = ctuple(C_THEN_ELSE, -1, block);
        return else_c;
        
      } else if (alt == 2) {     //\\   "else" <UI> <S>,
        int rest   = STATEMENT_BLOCK(compile(child(astp,2)));
        return ctuple(C_THEN_ELSE, -1, rest);

      } else if (alt == 3) {     //\\   <S>;
        return -1;
      }

    case AST_THEN_S:             //\\ P<THEN-S> =
      if (alt == 0) {            //\\   <OPT-then> "start" <S> <INTERNALS> <FINISH-S>,
        int block  = STATEMENT_BLOCK(compile(child(astp,4)));
        int finish = compile(child(astp,5));
        if (finish == -1) {
          return ctuple(C_THEN_ELSE, block, -1);
        }
        if (AstOP(finish) == C_THEN_ELSE) { // #1 "then" block (may be -1)  #2 "else" block (may be -1).
          assert(leftchild(finish) == -1);
          leftchild(finish) = block;
          // rightchild(finish) is any potential "else" part. Or -1 if none.
          return finish;
        }

      } else if (alt == 1) {     //\\   "then" <UI> <Opt-ELSE-S>;
        int then   = STATEMENT_BLOCK(compile(child(astp,2)));
        int else_c = compile(child(astp,3));
        if (else_c == -1) {
          return ctuple(C_THEN_ELSE, then, -1);
        }
        if (AstOP(else_c) == C_THEN_ELSE) { // #1 "then" block (may be -1)  #2 "else" block (may be -1).
          assert(leftchild(else_c) == -1);
          leftchild(else_c) = then;
          return else_c;
        }
        return ctuple(C_THEN_ELSE, then, else_c);
        
      }

    case AST_OPT_then:           //\\ P<OPT-then> = "then", ;
      return -1; // syntactic sugar

    case AST_ONE_SWITCH_DECL:    //\\ P<ONE-SWITCH-DECL> = <NAME> <Opt-NAME-LIST> '(' <CEXPR> ':' <CEXPR> ')'; // should be <CEXPR>s!!!
      if (alt == 0) {
        int name        = compile(child(astp,1)); assert(AstOP(name) == C_NAME);
        {
          char orig[1024];
          orig[0] = '\0'; getstr(child(astp,4), orig);
          //fprintf(stderr, "LOW: %s\n", orig);
          orig[0] = '\0'; getstr(child(astp,6), orig);
          //fprintf(stderr, "HIGH: %s\n", orig);
        }
        int lower_bound = compile(child(astp,4)); //fprintf(stderr, "lower bound: ");diagnose(lower_bound);
        assert(AstOP(lower_bound) == C_CONST_INT);
        int upper_bound = compile(child(astp,6)); //fprintf(stderr, "\nupper bound: ");diagnose(upper_bound);
        assert(AstOP(upper_bound) == C_CONST_INT);
        int boundspair  = ctuple(C_BOUNDSPAIR, lower_bound, upper_bound, lower_bound, upper_bound, -1);
        int more_names  = compile(child(astp,2)); assert(more_names == -1 || AstOP(more_names) == C_SEQ);
        int this_switch;
        int results = -1;
        int low         = rightchild(lower_bound); //fprintf(stderr, "low val: %d\n",low);
        int high        = rightchild(upper_bound); //fprintf(stderr, "high val: %d\n",high);
        
        for (;;) {
          this_switch = ctuple(C_SWITCH_DECL, name, boundspair, -1);
          //fprintf(stderr, "*** ADDING SWITCH %s\n", S(name));
          int switch_table = new_switch(pool_to_str(leftchild(name)), this_switch, low, high);
          child(this_switch,3) = switch_table;
          Declare(name, this_switch);
          append_to(&results, ctuple(C_SEQ, this_switch, -1));
          if (more_names == -1) break;
          detuple(more_names, C_SEQ, &name, &more_names);
        }
        return results;
      }

    //\\ #  rest of %switch list
    //\\ #  This reminds me, isn't there an optional %label declaration in Imp?  Haven't seen an example yet,
    //\\ #  I need to check the manuals to make sure I'm not confusing this with C...
    case AST_REST_OF_SWLIST:     //\\ P<REST-OF-SWLIST> =
      if (alt == 0) {            //\\   ',' <ONE-SWITCH-DECL> <REST-OF-SWLIST>,
        int next_switch   = compile(child(astp,2));
        int more_switches = compile(child(astp,3));
        return ctuple(C_SEQ, next_switch, more_switches);
      } else if (alt == 1) {     //\\   ;
        return -1;
      }
      break;

    //\\ #  the version without "for" is to support an old-style statement, %WHILE <cond> %THEN <ui>
    case AST_Percent_WU:         //\\ P<Percent-WU> =
      if (alt == 0) {            //\\   "while" <TOP-LEVEL-CONDITION>,
        int cond = compile(rightchild(astp));
        // #1 statement to be executed  #2 cond
        return ctuple(C_WHILE, -1 /* body */, cond);

      } else if (alt == 1) {     //\\   "until" <TOP-LEVEL-CONDITION>;
        int cond = compile(rightchild(astp));
        // #1 statement to be executed  #2 cond  #3 reserved for obscure and semi-illegal case
        //  of (%while <cond1> %cycle or %for <forctl> %cycle)/%repeat %until <cond2> ...
        return ctuple(C_UNTIL, -1 /* body */, cond /*ctuple(C_LOGNOT, cond) now reversed in out() */, -1);

      }

    case AST_Percent_WUF:        //\\ P<Percent-WUF> =
      if (alt == 0) {            //\\   "while" <TOP-LEVEL-CONDITION>,
        int cond = compile(rightchild(astp));
        // #1 statement to be executed  #2 cond
        return ctuple(C_WHILE, -1 /* body */, cond);

      } else if (alt == 1) {     //\\   "until" <TOP-LEVEL-CONDITION>,
        int cond = compile(rightchild(astp));
        // #1 statement to be executed  #2 cond  #3 reserved for obscure and semi-illegal case
        //  of (%while <cond1> %cycle or %for <forctl> %cycle)/%repeat %until <cond2> ...
        return ctuple(C_UNTIL, -1 /* body */, cond /*ctuple(C_LOGNOT, cond) now reversed in out() */, -1);

      } else if (alt == 2) {     //\\   "for" <NAME> <SCALAR-ASSIGN> <EXPR> ',' <EXPR> ',' <EXPR>;
        int name    = compile(child(astp, 2));
        int assop   = compile(child(astp, 3));
        int initial = compile(child(astp, 4));
        int step    = compile(child(astp, 6));
        int final   = compile(child(astp, 8));
        return ctuple(C_IMP_FOR, -1 /*body*/, name, assop, initial, step, final, IntConstant_c(initial), IntConstant_c(step), IntConstant_c(final));

      }

    case AST_Opt_CYCPARM:        //\\ P<Opt-CYCPARM> =
      if (alt == 0) {            //\\   <NAME> <SCALAR-ASSIGN> <EXPR> ',' <EXPR> ',' <EXPR>,
        int name    = compile(child(astp, 1));
        int assop   = compile(child(astp, 2));
        int initial = compile(child(astp, 3));
        int step    = compile(child(astp, 5));
        int final   = compile(child(astp, 7));
        return ctuple(C_IMP_FOR, -1 /*body*/, name, assop, initial, step, final, IntConstant_c(initial), IntConstant_c(step), IntConstant_c(final));
      } else if (alt == 1) {     //\\   ;
        return -1; // plain cycle start.
      }

    //\\ #  reversed conditions and 1-line cycles - UNTIL has to be handled differently TO DO
    case AST_REST_OF_SS1:        //\\ P<REST-OF-SS1> =
      if (alt == 0) {            //\\   <Percent-IU> <TOP-LEVEL-CONDITION> <S>,
        int if_or_unless = compile(leftchild(astp));
        int cond         = compile(rightchild(astp));
        if (AstOP(if_or_unless) == C_UNLESS) {
          AstOP(if_or_unless) = C_IF; // C_UNLESS should never percolate up to the top level...
          cond = ctuple(C_LOGNOT, cond); // Maybe a good place to apply De-Morgan?
        }
        assert(AstOP(if_or_unless) == C_IF);
        assert(leftchild(if_or_unless) == -1);
        assert(rightchild(if_or_unless) == -1);
        rightchild(if_or_unless) = cond;
        return if_or_unless;

      } else if (alt == 1) {     //\\   <Percent-WUF> <S>,
        return compile(leftchild(astp));

      } else if (alt == 2) {     //\\   <S>,
        return -1;

      } else if (alt == 3) {     //\\   ;
        return -1;
      }


  default:
    fprintf(stderr, "MISSING AST OPERATOR AST_%s IN compile_flowcontrol(@AST:%d)\n",
            AstOPName(astop),
            astp);
    return compile_inner(astp, extra_parameter, line);
  }
}

int generate_flowcontrol(int astp) {
  int castp;
  if (debug_flowcontrol) fprintf(stderr, "generate_flowcontrol(@AST:%d)\n", astp);
  castp = compile_flowcontrol(astp, 0);
  output_c_flowcontrol(castp);
}
