%endoflist  { do not output translated C code }

! THIS FILE *MUST* BE KEPT IN SYNCH WITH perms.h and perms.c

! These prims are a mixture of prims from multiple systems, in an attempt
! to be backward compatible with as many Imp environments as possible.
! Unfortunately a small number of library calls are incompatible between
! versions; specifically PRINT and PRINT FL which in one incarnation
! have one argument for places and in another have two.  It's marginally
! possible to support both using either stdargs, or the C11 support for 
! varadic macros.

! How it all works: this file does not cause any C code to be output - the
! C output only start on %endofperm ... this exists only to get the declarations
! into the symbol table for use by the program that follows.  When compiled
! by the C compiler, it uses perms.h instead, which affords us the opportunity
! to do some tricks with the C pre-processor etc which come in handy for
! handling signals for example.

! procedures declared as %perm or %prim are treated similarly to standard
! external routines, except that they are declared at an outer level of scope
! and can therefore be superceded by a user-written top-level procedure.

! %perm routines are just normal routines as far as the translation to C
! is concerned - but %prim routines are often implemented by code in the compiler
! itself and there may be no actual body for these procedures (although
! there *can* be, in addition to generated code). They can also be implemented
! by macro calls or C's "inline" procedures from within perm.h.

! Note also that %systemroutine procedures are *C* externals, not ones
! written in Imp - the primary difference (apart from trivial name issues)
! is that C procedures evaluate their parameters in right-to-left order
! as opposed to Imp which is defined as evaluating parameters from left-to-right.


%conststring(255)       BUG = "THIS IS SYM 0"
! I was having problems with unassigned variables in the compiler outputting Sym[x]
! where x was erroneously 0 ... so I created this variable as the first to be
! compiled, so that any references to Sym[0] (*or* stringpool[0]) would be obvious!
! Since the 'sym[]' array will be going away soon, so will this.

%constlongreal          PI = 3.141592653589793238462

! prims are implemented by the compiler with the equivalent of inline code

%primlongrealfnspec     float(%long %real n)  ;! not %integer parameter - relies on compiler promoting int to double when assigning to a %real
%primstring(1)%fnspec   tostring(%integer c)
%primintegerfnspec      addr(%name p)
%primintegermapspec     integer(%integer n)   ;! any parameter which has to represent an address should be %longinteger once it is supported.
                                              ;! and for compiling pre-64bit code, we probably need to invent a "%integers %long" statement!
%primbyteintegermapspec byteinteger(%integer n)
%primstring(*)%mapspec  string(%integer n)
%primrecord(*)%mapspec  record(%integer n)
%primrealmapspec        real(%integer n)
%primlongrealmapspec    longreal(%integer n)
%primbyteintegermapspec length(%string(255) s)
%primbyteintegermapspec charno(%string(255) s,%integer n)

%primlonglongintegerfnspec lengtheni(%integer i)

!integer function TYPE OF ( name X )
!
!This function returns a code which indicates the type of the
!object supplied as parameter.  The complete list of code values
!which may be returned by TYPE OF is as follows:
!
!0 - unknown type
!1 - integer
!2 - real
!3 - string
!4 - record
!5 - byte integer
!6 - short integer
!7 - long integer
!8 - long real
!9 - array
!10 - label

%primintegerfnspec        typeof(%name n)
%primintegerfnspec        sizeof(%name n)

%constinteger             nl = 10
%conststring(1)           snl = "
"
!%constname               nil == 0  ;! not sure if this syntax will work...


%systemlongrealfnspec     arcsin  %alias "asin"  (%longreal angle)
%systemlongrealfnspec     arccos  %alias "acos"  (%longreal angle)
%systemlongrealfnspec     arctan  %alias "atan"  (%longreal x,y)
%systemlongrealfnspec     arctan1 %alias "atan1" (%longreal angle)

%systemlongrealfnspec     sin(%longreal angle)
%systemlongrealfnspec     cos(%longreal angle)
%systemlongrealfnspec     tan(%longreal angle)

%permlongrealfnspec       fraction(%longreal r)
%permlongrealfnspec       frac pt(%longreal r)

%permintegerfnspec        imod(%integer i)
%permlongrealfnspec       mod(%longreal r)

%permintegerfnspec        int(%longreal r)
%permintegerfnspec        int pt(%longreal r)

%permlongintegerfnspec    lint(%longreal r)
%permlongintegerfnspec    lint pt(%longreal r)

%permintegerfnspec        trunc(%longreal r)
%permintegerfnspec        round(%longreal r)

%permlongintegerfnspec    iexp(%integer num, power)
%permlongrealfnspec       rexp(%longreal num, power)

%permlongrealfnspec       log(%longreal X)
%primintegerfnspec        rem(%integer p,q)
%systemlongrealfnspec     sqrt(%longreal num)
%systemintegerfnspec      isqrt(%longreal num)
%permintegerfnspec        mul div(%integer A, B, C )

! Perms are roughly equivalent to externals at level -1, however
! they *may* be implemented by macros.  Therefore Perms should not
! be assumed to usable as procedure parameters - if needed, the
! traditional method is to wrap them in another procedure.

%permroutinespec          prompt(%string(15) s)
%permroutinespec          read symbol(%integername p)  { *not* %byteintegername, which may cause problems }
%permroutinespec          read ch(%integername p)
%permintegerfnspec        next symbol  { 'symbol' does cr/lf filtering if appropriate }
%permintegerfnspec        next ch      { whereas 'ch's are raw and may return CR }
%permroutinespec          skip symbol
%permroutinespec          print symbol(%byteinteger sym)
%permroutinespec          print ch(%byteinteger sym)
%permroutinespec          print string(%string(255)  s)
%permroutinespec          write(%integer  v,p)

! "item"s are just 1-character strings.
%externalroutinespec      read item(%string(*)%name s)
%externalroutinespec      read string(%string(*)%name s)
%externalroutinespec      read text(%string(*)%name s,%integer delim)
%externalstring(255)%fnspec next item
%permroutinespec          read line(%string(*)%name s)

%permintegerfnspec        instream
%permintegerfnspec        outstream
%permintegerfnspec        input stream
%permintegerfnspec        output stream
%permstring(255)%fnspec   input name
%permstring(255)%fnspec   output name
%permstring(255)%fnspec   in file name
%permstring(255)%fnspec   out file name
%permroutinespec          select input(%integer  n)
%permroutinespec          select output(%integer  n)
%permroutinespec          open input(%integer  n, %string(255) fd)
%permroutinespec          open output(%integer  n,%string(255) fd)
%permroutinespec          open binary input(%integer  n, %string(255) fd)
%permroutinespec          open binary output(%integer  n,%string(255) fd)
%permroutinespec          define input(%integer i,%string(255) spec)
%permroutinespec          define output(%integer i,%string(255) spec)
%permroutinespec          abandon input
%permroutinespec          abandon output
%permroutinespec          close input
%permroutinespec          close output
%permroutinespec          reset input
%permroutinespec          reset output
%permroutinespec          complete input
%permroutinespec          complete output
%permroutinespec          position input(%integer p)
%permroutinespec          position output(%integer p)
%permroutinespec          input position
%permroutinespec          output position
%permpredicatespec        end of input
%permroutinespec          space
%permroutinespec          spaces(%integer  n)
%externalroutinespec      newpage
%permroutinespec          newline
%permroutinespec          newlines(%integer n)
%permroutinespec          read(%name ptr)
%permroutinespec          print(%longreal r, %integer before, %integer after)
%permroutinespec          print floating(%longreal r, %integer a, b)
%permroutinespec          print fl(%longreal r, %integer places)
%permroutinespec	  to upper(%string(*)%name s)
%permroutinespec	  to lower(%string(*)%name s)
%permstring(255)%fnspec   sub string(%string(255) s, %integer from, to)
%permstring(255)%fnspec  from string(%string(255) s, %integer from, to)
%permstring(255)%fnspec   trim(%string(255) s, %integer max)
! instream & outstream are currently just integers.  Rename them as _imp_ and return via these procedures below.
!%externalroutinespec      reset input
!%externalroutinespec      reset output
%externalstring(8)%fnspec time
%externalstring(9)%fnspec date
%externalintegerfnspec    cpu time
%externalstring(255)%fnspec cli param
%permintegerfnspec        event
%permintegerfnspec        subevent
%permintegerfnspec        eventinfo
%externalstring(255)%fnspec itos(%integer i,pos)

! Signal event codes
%constinteger PROGRAM STOP = 0
%constinteger OVERFLOW = 1
%constinteger STORE EXCEEDED = 2
%constinteger CONVERSION ERROR = 4
%constinteger ARGUMENT ERROR = 5
%constinteger RANGE ERROR = 6
%constinteger STRING RESOLUTION FAILS = 7
%constinteger UNDEFINED VALUE = 8
%constinteger INPUT ENDED = 9
%constinteger IO ERROR = 10
%constinteger JUMP OUT = 11
%constinteger USER EVENT = 12

%list { resume outputting generated C code }

%endofperm
